import warnings

# We can use UserWarning here instead to defeat the default filter but that is less informative
warnings.simplefilter("always", DeprecationWarning)
warnings.warn(
    "Import the ifboard from mkidgen3.equipment_drivers",
    DeprecationWarning,
    stacklevel=0,
)

from mkidgen3.equipment_drivers.ifboard import *



import threading
import logging
import time

import mkidgen3.registers as registers

from enum import Enum, IntFlag
from math import floor

from pynq import DefaultIP

NS_PER_SEC = 1000 * 1000 * 1000
SUBNS_PER_NS = 256


def _field_get(reg, lsb, mask):
    return (reg >> lsb) & mask


def _field_set(reg, lsb, mask, val):
    current = reg & (0xFFFFFFFF ^ (mask << lsb))
    return current | ((val & mask) << lsb)


def _make_time_ns(secs, ns, subns):
    return secs * NS_PER_SEC + ns + subns / SUBNS_PER_NS


class PPSMode(Enum):
    """Enumerate PPS synchronizer modes."""

    STOPPED = 0
    """Stops the PPS synchronization engine"""

    FORCE_START = 1
    """Immediately starts the PPS Engine ignoring PPS pulses"""

    PPS_FREERUN = 2
    """Starts on the next PPS edge and then counts ignoring all further PPS
    edges
    """

    PPS_SYNC = 3
    """Probably the most useful mode, starts on the next PPS edge and rolls
    over on subsequent PPS edges
    """

    SYSREF_SYNC = 4
    """Starts on the next SYSREF edge and rolls over on subsequent PPS edges,
    this is only really useful with the SYNC gated Sysref of the LMK04828 to
    sync boards to < 1 clock period
    """

    SYSREF_FREERUN = 5
    """Starts on the next SYSREF edge then lets the counter run free"""


class CaptureMode(Enum):
    """Enumerates counter capture modes."""

    CLEAR = 0
    """Clear the current capture and initilize the capture engine for another
    capture.
    """

    PPS_EDGE = 1
    """Capture the counter value on the next PPS Edge"""

    SYSREF_EDGE = 2
    """Capture the counter value on the next SYSREF Edge"""


class SanityMode(IntFlag):
    """Enumartes options for sanity checking on PPS rollover."""

    LOCKOUT = 1
    """Effectively debouncing, ignores PPS Edges for a period of time
    after the counter rolls over
    """

    SELF_ROLLOVER = 2
    """If the counter exceeds a certain threshold then rollover without a PPS
    Edge
    """

    RESYNC = 4
    """Handy for a flaky PPS signal that drops in and out; enables both LOCKOUT
    and MISSDETECT, but if a PPS edge is detected inside the debounce window
    then we disable the next MISSDETCT window and wait for the next pulse.

    If the PPS generator is really bad you probably want PPSMode.PPS_FREERUN
    and you can use the capture engine and live setting of the clock period
    to implement a software PLL.
    """


class PPSSource(IntFlag):
    """As of 9/13/23 the pps engine has three PPS inputs connected as described,
    they can be or-ed together to generate an internal PPS pulse by or-ing
    several of these enums together (IE `PPSSource.PPS0 | PPSSource.PPS1`)
    """

    PPS0 = 1
    """As of 9/13/23 this comes from the schmidt trigger on the RFSoC4x2"""

    PPS1 = 2
    """As of 9/13/23 this comes from the comparator on the RFSoC4x2"""

    PPS2 = 4
    """As of 9/13/23 this is unconnected on the RFSoC4x2, in the future it
    may be connected to the slow ADC that reads the PPS input"""


class PPSSync(DefaultIP):
    bindto = ["MazinLab:mkidgen3:pps_synchronizer_control:0.3"]

    def __init__(self, description):
        super().__init__(description=description)
        self.capture_event = threading.Event()
        self.started_event = threading.Event()

    # High-level api
    def start_engine(
        self,
        mode=PPSMode.PPS_SYNC,
        start_second=None,
        skew=None,
        lockout=None,
        rollover_thresh=None,
        resync=False,
        pps_source=PPSSource.PPS0,
        load_time=None,
        clk_period_ns=1.953125,
        timeout=5,
        poll=0,
    ):
        """Starts the PPS engine in varius modes

        This will configure the PPS engine and provides the tools needed to
        achieve board to board timestamp synchronization.

        If you have a reliable PPS Source with PPS edges on the second edge
        and assumptions about board clock accuracy are met then you can simply
        do `start_engine(skew = 0)` on all of the boards you want to
        synchronize.

        If your PPS Edges are not aligned then you will likely want go choose
        a board and measure the skew relative to the system clock using
        `PPSSync.sample_skew` and pass that skew to all the boards you want
        to start with `start_engine(skew = skew_measured)`. If you are
        using only one board then `start_engine()` will work

        If you do not have a PPS Source available then using
        `start_engine(PPSMode.FORCE_START)` will start the PPS Engine
        with a the system time plus or minus around a dozen microseconds

        Parameters
        ----------
        mode : PPSMode
            The mode you would like to start the PPS engine, SYSREF modes are
            not considered part of the stable API and should not be used unless
            you know what you are doing
        start_second : int or NoneType
            The UTC second you would like PPS engine to begin counting at,
            should be at least 4 seconds in the future or None
        skew : int or NoneType
            The skew between the true second and the PPS Signal edge
            in nanoseconds. If the PPS signal comes from a GPS source set this
            to zero otherwise set it to a common value accross all boards. If
            this is set to None the skew will be sampled relative to the system
            clock in PPS based modes.
        lockout : int or NoneType
            The debounce interval in nanoseconds during which PPS edges will be
            ignored. For the FS725 this can likely safely be left as None,
            otherwise set to the risetime of your PPS Signal.
        rollover : int or NoneType
            The number of nanoseconds after which the nanosecond counter should
            rollover without the presence of a PPSEdge. For the FS725 this can
            be safely set to 1000*1000*1000 if your clocking configuration
            is correct. If this and lockout are set resync should likely be set
        resync : bool
            If this is set and a PPS edge is detected in the lockout period
            after an automatic rollover then the counter will not rollover
            until another PPS Edge has been seen
        pps_source : PPSSource
            The PPS Source selected, see enum for documentation, PPS Signals
            can be ord together
        load_time : int or float or list or tuple or NoneType
            The time in nanoseconds (float or int), or a tuple/list of seconds,
            nanoseconds, and subns which you would like to load into the engine
            when it is started. If None the utc second on which we are starting
            will be chosen.
        clk_period_ns : float
            The clock period of the clock domain the block exists in. This
            should be 1/512MHz as of Nov 2023
        timeout : int or float
            The maximum number of seconds to wait for a pps edge before timing
            out worst case execution time is less than 4 + 2*timeout seconds
        poll : float
            The number of seconds to wait between polling for the status of
            the engine

        Assumptions
        -----------
        - The systems clock is accurate to < 0.2 seconds and that a leap
          second is not about to occur
        - The system is fast enough that ~12 writes to registers and
          a small amount of floating point math can be done in python in
          under a second
        """
        self.mode = PPSMode.STOPPED
        self.pps_source = pps_source
        self.delay_ns = 0
        self.started_event.clear()

        if lockout == 0:
            lockout = None

        if mode == PPSMode.PPS_SYNC or mode == PPSMode.PPS_FREERUN:
            if skew is None:
                skew = self.sample_skew(timeout, poll)
            if start_second is None:
                start_second = time.time() + 1
                if start_second % 1.0 > 0.75:
                    start_second = start_second + 1
                start_second = start_second // 1
        else:
            if skew is None:
                skew = 0
            if start_second is None:
                start_second = time.time()
        start_after = start_second - 0.2
        self.delay_ns = (NS_PER_SEC - skew) % NS_PER_SEC

        self.ns_per_clk = int(floor(clk_period_ns))
        self.subns_per_clk = int((clk_period_ns % 1.0) * 256)

        if load_time is None:
            load_time = int(start_second * NS_PER_SEC)
        elif type(load_time) == float:
            load_time = int(load_time)
        if type(load_time) == int:
            load_time = [load_time // NS_PER_SEC, load_time % NS_PER_SEC, 0]
        self.load_secs, self.load_ns, self.load_subns = list(load_time)

        sanity_mode = SanityMode(0)

        if lockout is not None:
            self.lockout = lockout
            sanity_mode |= SanityMode.LOCKOUT
        if rollover_thresh is not None:
            self.rollover_thresh = rollover_thresh
            sanity_mode |= SanityMode.SELF_ROLLOVER
        if resync:
            assert (
                sanity_mode == SanityMode.LOCKOUT | SanityMode.SELF_ROLLOVER
            ), "Resync requires both a lockout value and rollover threshold"
            sanity_mode |= SanityMode.RESYNC
        self.sanity_mode = sanity_mode

        if start_after is not None:
            while time.time_ns() < start_after:
                time.sleep(poll)

        self.mode = mode
        if mode == PPSMode.STOPPED:
            self.started_event.clear()
            return

        start_time = time.time()
        while timeout == 0 or time.time() - start_time < timeout:
            if self.started:
                self.started_event.set()
                return
            time.sleep(poll)
        raise TimeoutError(
            "PPS Engine did not start before timeout, is the PPS source correct?"
        )

    def stop_engine(self):
        """Stop the PPS Engine"""
        self.started_event.clear()
        self.mode = PPSMode.STOPPED

    def capture(
        self, capture_mode=CaptureMode.PPS_EDGE, timeout=5, poll=0, fmt=float
    ):
        assert fmt in (
            tuple,
            float,
        ), "Valid capture timestamp formats are tuple and float"
        self.capture_mode = CaptureMode.CLEAR
        self.capture_event.clear()
        self.capture_mode = capture_mode
        start_time = time.time()
        while timeout == 0 or time.time() - start_time < timeout:
            if self.captured:
                self.capture_event.set()
                if fmt == tuple:
                    return (self.captured_secs, self.captured_ns, self.captured_subns)
                if fmt == float:
                    return self.captured_time_ns
            time.sleep(poll)
        raise TimeoutError("Capture didn't occur before timeout")

    def sample_skew(self, timeout=5, poll=0):
        """Captures the sub-second portion of the system time at a PPS Edge
        The PPS Engine should be stopped before you call this. Should be
        accurate to less than 50 microseconds with the system under reasonable
        load.

        Parameters
        ----------
        timeout : int or float
            The amount of time to wait for a PPS Edge in seconds
        poll : float
            The time between checking for a capture in seconds

        Returns
        -------
        skew : int
            The skew between the PPS Edge and the system time in nanoseconds
        """
        assert (
            self.mode == PPSMode.STOPPED and self.delay_ns == 0
        ), "Turn off the counter, and set the delay to zero"
        self.start_engine(PPSMode.PPS_SYNC, timeout, poll)
        pps_time = time.time_ns()
        self.stop_engine()
        return pps_time % NS_PER_SEC

    # Base Registers
    @registers.register_ro
    def _counter_status_reg(self):
        return self.register_map.counter_status_reg

    @registers.register_shadow(0)
    def _mode_reg(self):
        """Provides a read write interface to the underlying mode register
        which is write-only, should not be called by user code as _mode_reg
        contains several config registers one of which is also confusingly
        what is set by PPSSync.mode, I'm very sorry."""
        return self.register_map.mode_reg

    @registers.register_shadow(0)
    def _counter_config_reg(self):
        return self.register_map.counter_config_reg

    # Quasi user-accessible registers
    @registers.register_shadow(0)
    def lockout(self):
        """The lockout interval in ns (See :py:class:`mkidgen3.drivers.ppssync.SanityMode`)"""
        return self.register_map.lockout_reg

    @registers.register_shadow(0)
    def rollover_thresh(self):
        """The rollover threshold in ns (See :py:class:`mkidgen3.drivers.ppssync.SanityMode`)"""
        return self.register_map.rollover_thresh_reg

    @registers.register_shadow(0)
    def load_secs(self):
        """The seconds counter value to load when we start the PPS engine"""
        return self.register_map.load_secs_reg

    @registers.register_shadow(0)
    def load_ns(self):
        """The ns counter value to load when we start the PPS engine"""
        return self.register_map.load_ns_reg

    @registers.register_shadow(0)
    def load_subns(self):
        """The subns counter value to load when we start the PPS engine"""
        return self.register_map.load_subns_reg

    @registers.register_shadow(0)
    def delay_ns(self):
        """The time in ns to delay each PPS Edge"""
        return self.register_map.delay_ns_reg

    @registers.register_ro
    def captured_subns(self):
        """The raw captured subns counter value"""
        return self.register_map.capture_subns_reg

    @registers.register_ro
    def captured_ns(self):
        """The raw captured ns counter value"""
        return self.register_map.capture_ns_reg

    @registers.register_ro
    def captured_secs(self):
        """The raw captured secs counter value"""
        return self.register_map.capture_secs_reg

    @registers.register_ro
    def subns(self):
        """The raw current subns counter value"""
        return self.register_map.current_subns_reg

    @registers.register_ro
    def ns(self):
        """The raw current ns counter value"""
        return self.register_map.current_ns_reg

    @registers.register_ro
    def secs(self):
        """The raw current secs counter value"""
        return self.register_map.current_secs_reg

    # Individual register fields, quasi user-accessible    
    started = registers.field_bool(0, _counter_status_reg)
    captured = registers.field_bool(8, _counter_status_reg)

    mode = registers.field_enum(slice(0, 4), PPSMode, _mode_reg)
    capture_mode = registers.field_enum(slice(8, 10), CaptureMode, _mode_reg)
    sanity_mode = registers.field_enum(slice(16, 19), SanityMode, _mode_reg)
    pps_source = registers.field_enum(slice(24, 27), PPSSource, _mode_reg)

    ns_per_clk = registers.field(slice(8, 14), _counter_config_reg)
    subns_per_clk = registers.field(slice(0, 8), _counter_config_reg)

    # Timer readout stuffs
    @property
    def captured_time_ns(self):
        """The time captured by the core in floating point nanoseconds"""
        if not self.captured:
            logging.getLogger(__name__).warning(
                "Captured time read while registers are invalid"
            )
        return _make_time_ns(self.captured_secs, self.captured_ns, self.captured_subns)

    @property
    def captured_time(self):
        """The Time captured by the core in floating point seconds"""
        return self.captured_time_ns / (1000 * 1000 * 1000)

    @property
    def current_time_ns(self):
        """The current time according to the core in nanoseconds"""
        secs = self.secs
        # subns just won't be accurate
        ns = self.ns
        after_secs = self.secs
        if after_secs != self.secs:
            return self.current_time_ns
        return _make_time_ns(secs, ns, 0)

    @property
    def current_time(self):
        """The current time according to the core in floating point seconds"""
        return self.current_time_ns / (1000 * 1000 * 1000)



import zmq
import os
import uuid


def zpipe(ctx):
    """
    build an inproc pipe for talking to threads
    mimic pipe used in czmq zthread_fork.
    Returns a pair of PAIRs connected via inproc
    """
    a = ctx.socket(zmq.PAIR)
    b = ctx.socket(zmq.PAIR)
    a.linger = b.linger = 0
    a.hwm = b.hwm = 1
    iface = "inproc://%s" % uuid.uuid4().hex
    a.bind(iface)
    b.connect(iface)
    return a, b



import copy

from mkidgen3.server.feedline_config import *
from mkidgen3.server.waveform import WaveformFactory
from mkidgen3.opfb import opfb_bin_number

if1 = IFConfig(lo=6.0, adc_attn=(3.0, None), dac_attn=None)
if1a = IFConfig(lo=6.0, adc_attn=(3.0, None), dac_attn=3.0)
if2 = IFConfig(lo=3.0, adc_attn=(3.0, None), dac_attn=3.0)
if3 = IFConfig(lo=6.0, adc_attn=(3.0, None), dac_attn=None)

# Bitstream Config
bitstream = BitstreamConfig(bitstream='/home/xilinx/gen3_top.bit', ignore_version=True)
rfdc_clk = RFDCClockingConfig(programming_key='4.096GSPS_MTS_dualloop', clock_source=None)  # clock source should default to external 10 MHz
rfdc = RFDCConfig(dac_mts=True, adc_mts=False, adc_gains=None, dac_gains=None)
if_board = IFConfig(lo=6000, adc_attn=10, dac_attn=50)
waveform_vals = WaveformFactory(frequencies=[100e6])
waveform = WaveformConfig(waveform=waveform_vals)

waveform_vals2 = WaveformFactory(frequencies=[150e6])
waveform2 = WaveformConfig(waveform=waveform_vals2)
freqs = waveform.waveform.freqs

# Bin2Res Config
bins = np.zeros(2048, dtype=int)
bins[:freqs.size] = opfb_bin_number(freqs, ssr_raw_order=True)
chan = ChannelConfig(bins=bins)

# DDC Config
ddc_tones = np.zeros(2048)
ddc_tones[:freqs.size]=freqs
ddc = DDCConfig(tones=ddc_tones)

# Feedline Config
fc = FeedlineConfig(bitstream=bitstream, rfdc_clk=rfdc_clk, rfdc=rfdc,
                    if_board=if_board, waveform=waveform, chan=chan, ddc=ddc)
fc2 = FeedlineConfig(bitstream=bitstream, rfdc_clk=rfdc_clk, rfdc=rfdc,
                    if_board=if_board, waveform=waveform2, chan=chan, ddc=ddc)


assert fc2.compatible_with(fc2.hashed_form)

# test hashed comparison
fc3 = copy.deepcopy(fc2)
fc3.if_board = IFConfig(lo=3000, dac_attn=50, adc_attn=50)

m = FeedlineConfigManager()

assert if1 < if1a and if1.compatible_with(if1a)
assert if1 < if2 and not if1.compatible_with(if2) and not if1a.compatible_with(if2)
assert if3.hashed_form == if1 and if3.hashed_form.compatible_with(if1)
assert if3.hashed_form != if1a and not if3.hashed_form.compatible_with(if1a)

assert fc3.compatible_with(fc3.hashed_form)


m.learn(fc)
assert m.unlearned_hashes(fc) == set()


update = m.add('1', fc)
assert update == fc
assert m.required() == fc
assert m.pop('1') == True

assert m.unlearned_hashes(fc2) == set()
#assert m.unlearned_hashes(x1a.hashed_form) == set([hash(if1a)])
#m.add('1', x1)
change = m.add('1a', fc2)
print(change)



import json
import threading
import math
import serial
import time
import numpy as np

from logging import getLogger
from typing import Tuple, List, Optional
from enum import IntFlag, auto
from secrets import token_bytes
from dataclasses import dataclass
from fractions import Fraction

from mkidgen3.registers import MetaRegister, field, field_bool, field_enum

MAX_OUT_ATTEN = 31.75  # dB
MAX_IN_ATTEN = 31.75  # dB
IF_ATTN_STEP = 0.25  # dB IF attenuator step size (minimum dB attenuation step, limited by a single attenuator)

def _escape_nline_creturn(x):
    """Escape \n and \r in a string"""
    return x.replace('\n', '\\n').replace('\r', '\\r')


class SerialDevice:
    def __init__(self, port, baudrate=115200, timeout=0.1, name=None, terminator='\n', response_terminator=''):
        self.ser = None
        self.port = port
        self.baudrate = baudrate
        self.timeout = timeout
        self.name = name if name else self.port
        self.terminator = terminator
        self._response_terminator = response_terminator
        self._rlock = threading.RLock()

    def _preconnect(self):
        """
        Override to perform an action immediately prior to connection.
        Function should raise IOError if the serial device should not be opened.
        """
        pass

    def _postconnect(self):
        """
        Override to perform an action immediately after connection. Default is to sleep for twice the timeout
        Function should raise IOError if there are issues with the connection.
        Function will not be called if a connection can not be established or already exists.
        """
        pass

    def _predisconnect(self):
        """
        Override to perform an action immediately prior to disconnection.
        Function should raise IOError in the event of any error.
        """
        pass

    def connected(self):
        return True if self.ser is not None and self.ser.isOpen() else False

    def connect(self, reconnect=False, raise_errors=True):
        """
        Connect to a serial port. If reconnect is True, closes the port first and then tries to reopen it. First asks
        the port if it is already open. If so, returns nothing and allows the calling function to continue on. If port
        is not already open, first attempts to create a serial.Serial object and establish the connection.
        Raises an IOError if the serial connection is unable to be established.
        """
        if reconnect:
            self.disconnect()

        try:
            if self.ser.isOpen():
                return
        except Exception:
            pass

        log=getLogger(__name__).getChild('io')
        log.debug(f"Connecting to {self.port} at {self.baudrate}")
        try:
            self._preconnect()
            self.ser = serial.Serial(port=self.port, baudrate=self.baudrate, timeout=self.timeout)
            self._postconnect()
            log.debug(f"port {self.port} connection established")
            return True
        except (serial.SerialException, IOError) as e:
            log.error(f"Connecting to port {self.port} failed: {str(e)}, will attempt disconnect.")
            if self.ser is not None:
                self.disconnect()
            if raise_errors:
                raise e
            return False

    def disconnect(self):
        """
        First closes the existing serial connection and then sets the ser attribute to None. If an exception occurs in
        closing the port, log the error but do not raise.
        """
        try:
            self._predisconnect()
            self.ser.close()
        except Exception as e:
            getLogger(__name__).getChild('io').info(f"Exception during disconnect: {e}")
        self.ser = None

    def format_msg(self, msg: str):
        """Subclass may implement to apply hardware specific formatting"""
        if msg and msg[-1] != self.terminator:
            msg = msg + self.terminator
        return msg.encode('utf-8')

    def send(self, msg: str, connect=True):
        """
        Send a message to  the serial device and get a response, optionally connecting as needed.
        Formats message according to `self.format_msg()` function before sending.

        Serial/IO errors will be raised. An IOError will be raised if the device does not indicate success.

        In the event of an IOError, disconnect from the port before logging and raising the error.

        Args:
            msg: the message to send
            connect: Try to connect before sending, if not set an attempt to send will raise an IOError

        Returns: the response from the board, possibly ''

        Raises: IOError in the event of a failure.

        """
        with self._rlock:
            if connect:
                self.connect()
            try:
                msg = self.format_msg(msg)
                getLogger(__name__).getChild('io').debug(f"Sending '{msg}'")
                self.ser.write(msg)
                resp, success = self.receive()
                if not success:
                    raise IOError(resp)
                return resp
            except (serial.SerialException, IOError) as e:
                self.disconnect()
                getLogger(__name__).error(f"...failed: {e}")
                raise e

    def receive(self):
        """
        Receive data until a timeout or a :/n or ?/n is received.
        received, decode it and strip it of the confirmation/rejection. In the case of a serialException,
        disconnects from the serial port and raises an IOError.

        returns a tuple any respponse (minus the confiramtion/rejection) and a boolean indicating
        confirmation/rejection
        """
        with self._rlock:
            try:
                lines = []
                more = True
                while more:
                    lines.append(self.ser.readline().decode("utf-8"))
                    if lines[-1].endswith('?\r\n') or lines[-1].endswith(':\r\n'):
                        more = False
                response = ''.join(lines)
                getLogger(__name__).getChild('io').debug(f"Read {_escape_nline_creturn(response)} from {self.name}")
                return response.strip()[:-1], response.endswith(':\r\n')
            except (IOError, serial.SerialException) as e:
                self.disconnect()
                getLogger(__name__).debug(f"Send failed {e}")
                raise IOError(e)


class IFBoard(SerialDevice):
    def __init__(self, port='/dev/ifboard', timeout=0.1, connect=True):
        super().__init__(port, 115200, timeout, response_terminator='\r\n')
        self.firmware = None
        self._monitor_thread = None
        self._initialized = False
        self.initialized_at_last_connect = False
        self.rebooted = None
        self.trf_control = TRF3765((lambda a: self._trf_get_reg(a), lambda a, v: self._trf_set_reg(a, v)), 10)
        if connect:
            self.connect(raise_errors=False)

    def __str__(self):
        con= 'Connected' if self.connected() else 'Disconnected'
        return f'{self.port}@{self.baudrate} ({con})'

    def settle(self, timeout=None):
        """
        Wait for the IF board to be fully settled
        Returns: None
        """
        # Per Aled/Jenny 150us is the maximum settling time
        time.sleep(.000150)

    def set_lo(self, freq, fractional: Optional[bool] = None):
        if type(freq) is TRFCalibrationCertificate:
            assert fractional is None
            self.trf_control.set_output(freq, wait_for_lock=False)
        else:
            self.trf_control.set_output(freq / 2, wait_for_lock=False, fractional = fractional)

    def set_attens(self, output_attens: (float, Tuple[float], List[float], None) = None,
                   input_attens: (float, Tuple[float], List[float], None) = None):
        """
        Set attenuator values. Values are set in the order of the signal path. None implies a value is not changed.
        A single number sets both attenuation values.

        If individual attenuations are specified they are passed directly to the board without alteration.
        It is intended for low-level debugging use. Generally if setting individual attens,
        the majority of the attenuation should be in the first attenuator for the DAC and the second for the ADC.

        Allowed values are documented in the ifboard arduino codebase or the attenuator chips datasheet.

        Raises IOError on com failure, ValueError on invalid value or setting failure.
        """
        attens = self.attens
        current = [attens[v] for v in ('dac1', 'dac2', 'adc1', 'adc2')]

        if not isinstance(input_attens, (tuple, list)):
            x = min(input_attens, 2 * MAX_IN_ATTEN)

            if x!=input_attens:
                getLogger(__name__).warning(f'Max output attenuation exceeded. Clipping each value to {MAX_IN_ATTEN} dB.')

            input_attens = [max(x - MAX_IN_ATTEN, 0), min(x, MAX_IN_ATTEN)]

        if len(input_attens) != 2:
            raise ValueError('Incorrect number of input attenuations')

        if not isinstance(output_attens, (tuple, list)):
            x = min(output_attens, 2 * MAX_OUT_ATTEN)

            if x != output_attens:
                getLogger(__name__).warning(f'Max input attenuation exceeded. Clipping each value to {MAX_OUT_ATTEN} dB.')

            output_attens = [min(x, MAX_OUT_ATTEN), max(x - MAX_OUT_ATTEN, 0)]

        if len(output_attens) != 2:
            raise ValueError('Incorrect number of output attenuations')

        if (((np.asarray(input_attens) % IF_ATTN_STEP).any() != 0) or ((np.asarray(output_attens) % IF_ATTN_STEP).any() != 0)).any():
            getLogger(__name__).warning(f'Exact attenuation not achievable. Clipping value to nearest {IF_ATTN_STEP} dB. Check status() for exact value.')

        new = output_attens + input_attens
        try:
            attens = [f'{float(n if n is not None else c):.2f}' for n, c in zip(new, current)]
        except TypeError:
            raise ValueError('Attenuations must be float or None')

        self.send('AT' + ','.join(attens))

    @property
    def powered(self):
        powered = self.send('IO?')
        if powered == '1':
            return True
        elif powered == '0':
            return False
        else:
            raise IOError("Board did not adhere to protocol")

    @property
    def attens(self):
        """
        Returns a dict of attenuator values in dB: adc1, adc2, dac1, dac2

        These values are either in effect or will take effect when powered
        """
        atten = self.send('AT?')
        try:
            attens = json.loads(atten)
        except json.JSONDecodeError:
            raise IOError("IF board did not adhere to protocol")
        return attens

    def _postconnect(self):
        data = self.ser.readall().decode('utf-8')
        if 'Enter PC for cmd list' in data:
            self.rebooted = True
            self.ser.write('x\n'.encode('utf-8'))  # a throwaway string just to stop any boot message
        self.ser.readall()

    @property
    def lo_freq(self):
        """Get the LO frequency in use while powered"""
        return float(self.send('LO?'))

    def save_settings(self):
        """Save current settings as power-on defaults"""
        self.send('SV')

    def power_off(self, save_settings=True):
        """Turn off IF board power, optionally saving the active settings"""
        self.send('IO0' + ('S' if save_settings else ''))

    def power_on(self):
        """Power up and load saved settings"""
        self.send('IO1')

    def raw(self, cmd):
        """Send a raw serial command to the IF board and return the response"""
        return self.send(cmd)

    def status(self):
        response = self.send('TS')
        getLogger(__name__).debug(response)
        lines = response.split('\n')
        lines = [l for l in lines if not l.startswith('#')]
        return IFStatus(''.join(lines))

    def stop_lo(self):
        """
        Stop the LO from oscillating
        Returns:

        """
        getLogger(__name__).warning('Requested to stop LO, but function is presently NOOP')
        pass

    def configure(self, lo=None, dac_attn=None, adc_attn=None):
        getLogger(__name__).info(f'Configuring {self} to lo={lo} DAC Atten:{dac_attn} ADC Atten: {adc_attn}')
        self.power_on()
        self.set_lo(lo)
        self.set_attens(dac_attn, adc_attn)

    def _trf_get_reg(self, address):
        resp = self.send("RR{:d}".format(address))
        return int(resp.strip(), 16)

    def _trf_set_reg(self, address, value):
        self.send("WR{:d}{:s}".format(address, hex(value)[2:]))

class IFStatus:
    def __init__(self, jsonstr):
        self._data = d = json.loads(jsonstr)
        self.general = g = d['global']
        self.fresh_boot = not g['coms']
        self.boot = g['boot']
        self.power = g['power']
        self.trf_general = t = d['trf'][0]
        self.trf_regs = d['trf'][1:]
        self.dac_attens = (d['attens']['dac1'], d['attens']['dac2'])
        self.adc_attens = (d['attens']['adc1'], d['attens']['adc2'])
        partialcal = not (g['gen2'] or g['g3fcal'])
        s = 'LO gen{} {} mode, {} calibration. PLL {}locked.\n\tReq: {} MHz Attained: {} MHz Err: {} MHz'
        self.lo_mode = s.format('32'[g['gen2']], ('integer', 'fractional')[g['fract']], ('full', 'partial')[partialcal],
                                ('un', '')[t['pll_locked']], g['lo'], t['f_LO'], (t['f_LO'] or np.nan) - g['lo'])

    def __str__(self):
        stat = 'IFStatus: {}, boot {}{}. {}\n\tDAC attens: {}\n\tADC Attens: {}'
        return stat.format('Powered' if self.power else 'Unpowered', self.boot,
                           ', freshly booted' if self.fresh_boot else '',
                           self.lo_mode, self.dac_attens, self.adc_attens)


# It's ugly but its the best we can reasonably do atm
class _TRFReg(MetaRegister):
    cacheacc = 0
    totalacc = 0
    readacc = 0
    readcacc = 0
    def __init__(self, address, rwhandle, cached=True, nocache=0, readcache=False):
        self.address = address
        self.rwhandle = rwhandle
        self.cached = cached
        self.nocache = 0
        self.cache = None
        self.readcache = readcache

    def __len__(self):
        return 32

    def __get__(self, obj, objtype=None):
        if obj is None or ((objtype is not None) and issubclass(objtype, MetaRegister)):
            return self
        self.readacc += 1
        if self.readcache and self.cache is not None:
            self.readcacc += 1
            return self.cache
        self.cache = self.rwhandle[0](self.address)
        return self.cache

    def __set__(self, obj, val: int):
        self.totalacc += 1
        if self.cache is not None:
            if self.cache == val and self.cached and not (val & self.nocache):
                self.cacheacc += 1
                return
        self.cache = val
        self.rwhandle[1](self.address, val)

    def invalidate(self):
        self.cache = None

class TRFPowerDown(IntFlag):
    PLL = auto()
    CP = auto()
    VCO = auto()
    VCOMUX = auto()
    DIV124 = auto()
    PRESC = auto()
    LO_DIV = auto()
    BUFF1 = auto()
    BUFF2 = auto()
    BUFF3 = auto()
    BUFF4 = auto()

@dataclass
class TRFVCOCalibration:
    vco_sel: int
    vco_trim: int

@dataclass
class TRFFractionalCalibration:
    mod_ord: int
    isource_sink: bool
    isource_trim: int

DEFAULT_FRACCAL = TRFFractionalCalibration(0b10, False, 0b100)

@dataclass
class TRFDividerConfig:
    lo_div_sel: int
    pll_div_sel: int
    rdiv: int
    nint: int
    nfrac: int
    prsc_sel: int
    f_ref: int | Fraction
    
    def __post_init__(self):
        self.validate(False)

    def validate(self, calibrate: bool):
        if self.prsc_sel < 0 or self.prsc_sel > 1:
            raise ValueError("PRSC_SEL must be 0 or 1")
        if self.rdiv < 1 or self.rdiv > (1 << 13) - 1:
            raise ValueError("RDIV is not in [1, 1 << 13 - 1]")
        if self.nint < 1 or self.nint > (1 << 16) - 1:
            raise ValueError("NINT is not int [1, 1 << 16 - 1]")
        if self.nfrac < 0 or self.nfrac > (1 << 25) - 1:
            raise ValueError("NFRAC is not in [0, 1 << 25 - 1]")
        if self.pll_div_sel < 0 or self.pll_div_sel > 2:
            raise ValueError("PLL_DIV_SEL must be 0 1 or 2")
        if self.lo_div_sel < 0 or self.lo_div_sel > 3:
            raise ValueError("LO_DIV_SEL must be 0 1 2 or 3")

        if self.frequency > 4800 or self.frequency < 300:
            raise ValueError("Frequency {:f} out of range [300, 4800] MHz".format(float(self.frequency)))
        if self.f_vco / self.pll_div > 3000:
            raise ValueError("PLL Input frequency {:f} > 3000 MHz".format(float(self.f_vco / self.pll_div)))
        if self.f_pfd > 48:
            raise ValueError("Phase detector frequency ({:f} MHz) > 48 MHz".format(float(self.f_pfd)))
        if self.f_ref < 0.5 or self.f_ref > 350:
            raise ValueError("Reference frequency ({:f} MHz) < 0.5MHz or < 350 MHz".format(float(self.f_ref)))
        if self.f_n > 375:
            raise ValueError("Integer divider frequency (f_n = {:f}) > 375 MHz".format(float(self.f_n)))

        if self.prsc_p == 8:
            if self.nfrac == 0 and self.nint < 72:
                raise ValueError("PRSC is 8/9 with NINT < 72 in integer mode")
            elif self.nfrac != 0 and self.nint < 75:
                raise ValueError("PRSC is 8/9 with nint < 75 in fractional mode")
        if self.prsc_p == 4:
            if self.nfrac == 0 and (self.nint >= 72 or self.nint < 20):
                raise ValueError("PRSC is 4/5 with nint not in [20, 72)")
            if self.nfrac == 1 and (self.nint >= 75 or self.nint < 23):
                raise ValueError("PRSC is 4/5 with nint not in [23, 75)")

        if calibrate:
            _ = self.cal_clk

    @property
    def frequency(self) -> Fraction:
        return self.f_vco / self.lo_div

    @property
    def step(self) -> Fraction:
        return TRFDividerConfig(
            self.lo_div_sel,
            self.pll_div_sel,
            self.rdiv,
            self.nint,
            self.nfrac + 1,
            self.prsc_sel,
            self.f_ref
        ).frequency - self.frequency

    @property
    def prsc_p(self) -> int:
        if self.prsc_sel:
            return 8
        return 4

    @property
    def pll_div(self) -> int:
        return 1 << self.pll_div_sel

    @property
    def lo_div(self) -> int:
        return 1 << self.lo_div_sel

    @property
    def f_vco(self) -> Fraction:
        base = Fraction(self.f_ref * self.pll_div, self.rdiv)
        integer = base*self.nint
        fractional = base*Fraction(self.nfrac, 1<<25)
        return integer + fractional

    @property
    def f_pfd(self) -> Fraction:
        return self.f_vco / (self.nint * self.pll_div)

    @property
    def f_n(self) -> Fraction:
        return self.f_vco / (self.prsc_p * self.pll_div)

    @property
    def cal_clk_factor(self) -> Fraction:
        pfd = self.f_pfd
        ref = self.f_ref
        factor = Fraction(128, 1)
        while pfd * factor > Fraction(6, 10):
            factor = factor / 2
            if factor < Fraction(1, 128):
                raise ValueError("Unable to produce a cal_clk frequency < 600 KHz")
        while ref * factor > Fraction(1, 100) and pfd * factor > Fraction(5, 100) and factor >= Fraction(1, 128):
            if ref / (pfd * factor) < 8000:
                return factor
            factor = factor / 2
        raise ValueError("Unable to find valid cal clk")

    @property
    def cal_clk(self) -> Fraction:
        return self.cal_clk_factor * self.f_pfd

    @property
    def cal_clk_sel(self) -> int:
        n, d = self.cal_clk_factor.as_integer_ratio()
        if n == d and n == 1:
            return 0b1000
        if n == 1:
            return 0b1000 + round(math.log2(d))
        return 0b1000 - round(math.log2(n))

    @classmethod
    def from_target(cls, freq_unscaled: float | Fraction, f_ref: int | Fraction = 10) -> "TRFDividerConfig":
        if freq_unscaled is float:
            freq = Fraction(freq_unscaled)
        else:
            freq = freq_unscaled

        if freq > 4800 or freq < 300:
            raise ValueError("TRF3765 (undoubled) output frequency not in [300, 4800] MHz")
        if freq >= 2400:
            lo_div_sel = 0
        elif freq >= 1200:
            lo_div_sel = 1
        elif freq >= 600:
            lo_div_sel = 2
        else:
            lo_div_sel = 3
        f_vco_target = freq * (1 << lo_div_sel)

        pll_div_sel = 0
        while f_vco_target / (1 << pll_div_sel) > 3000:
            pll_div_sel += 1
            if pll_div_sel > 2:
                raise ValueError(
                    "Unable to compute pll_div_sel (see TRF3765 datasheet section 7.4.4.1 b ) with lo_div_sel = {:d}, freq_unscaled = {:f}, f_ref = {:f}".format(
                        lo_div_sel,
                        float(freq_unscaled),
                        float(f_ref)
                    )
                )

        rdiv = 1
        if f_ref >=  48:
            rdiv = math.ceil(48 / f_ref)
            if rdiv > (1 << 13) - 1:
                raise ValueError("Reference out of range, cannot make f_pfd <= 48MHz using rdiv")

        while pll_div_sel <= 2:
            pll_div = 1 << pll_div_sel
            nint = math.floor(f_vco_target * rdiv / (f_ref * pll_div))
            nfrac = round((f_vco_target * rdiv / (f_ref * pll_div) - nint) * (1 << 25))

            if nint >= 75:
                prsc_sel = 1
                p = 8
            elif nint >= 23:
                prsc_sel = 0
                p = 4
            else:
                pll_div_sel += 1
                continue

            f_n = f_vco_target / (p * pll_div)
            if f_n <= 375:
                return TRFDividerConfig(
                    lo_div_sel,
                    pll_div_sel,
                    rdiv,
                    nint,
                    nfrac,
                    prsc_sel,
                    f_ref
                )
        raise ValueError("Couldn't make digital divider frequency (f_N) <= 375 MHz, see TRF3765 Datasheet section 7.4.4.1 c.")

@dataclass
class TRFCalibrationCertificate:
    token: int
    divider_config: TRFDividerConfig
    vco_calibration: TRFVCOCalibration

    @property
    def frequency(self) -> Fraction:
        return self.divider_config.frequency

class TRF3765:
    tcal_total = 0.0

    def __init__(self, rwhandle, f_ref, outputs = TRFPowerDown.BUFF1):
        self.r0 = _TRFReg(0, rwhandle, cached=False)
        self.r1 = _TRFReg(1, rwhandle, readcache=True)
        self.r2 = _TRFReg(2, rwhandle, nocache=1<<31, readcache=True)
        self.r3 = _TRFReg(3, rwhandle, readcache=True)
        self.r4 = _TRFReg(4, rwhandle, readcache=True)
        self.r5 = _TRFReg(5, rwhandle, readcache=True)
        self.r6 = _TRFReg(6, rwhandle, readcache=True)

        self.__token = int.from_bytes(token_bytes(16), byteorder='big')
        self.f_ref = f_ref
        self.outputs = outputs

    def set_output(self, frequency: float | Fraction | TRFCalibrationCertificate, wait_for_lock: bool = True, fractional: Optional[bool] = None):
        if isinstance(frequency, TRFCalibrationCertificate):
            if frequency.token != self.__token:
                raise ValueError("Calibration certificate is not for this LO")
            self.__program_divider(frequency.divider_config)
            self.vco_sel = frequency.vco_calibration.vco_sel
            self.vco_trim = frequency.vco_calibration.vco_trim
        else:
            divider_config = TRFDividerConfig.from_target(frequency, self.f_ref)
            self.__program(divider_config, True, True, fractional)     
        if wait_for_lock:
            raise ValueError("Currently unimplemented")

    def get_certificate(self) -> TRFCalibrationCertificate:
        return TRFCalibrationCertificate(
            self.__token,
            TRFDividerConfig(
                self.lo_div_sel,
                self.pll_div_sel,
                self.rdiv,
                self.nint,
                self.nfrac,
                self.prsc_sel,
                self.f_ref
            ),
            TRFVCOCalibration(
                self.vco_sel_readback,
                self.vco_trim_readback,
            )
        )

    def __program_divider(self, divider_config: TRFDividerConfig):
        self.lo_div_sel = divider_config.lo_div_sel
        self.pll_div_sel = divider_config.pll_div_sel
        self.rdiv = divider_config.rdiv
        self.nint = divider_config.nint
        self.nfrac = divider_config.nfrac
        self.prsc_sel = divider_config.prsc_sel

    def __program(
        self,
        divider_config: TRFDividerConfig,
        calibrate: bool = True,
        powerdown: bool = True,
        fractional: Optional[bool] = None,
        fractional_cal: Optional[TRFFractionalCalibration] = None,
    ):
        assert divider_config.f_ref == self.f_ref
        if calibrate:
            self.cal_clk_sel = divider_config.cal_clk_sel
        if fractional is not None and not fractional:
            if divider_config.nfrac != 0:
                raise ValueError("Requested integer calibration with nfrac != 0")
        if fractional is None:
            fractional = divider_config.nfrac != 0
        self.powerdown_parts = TRFPowerDown.BUFF1 | TRFPowerDown.BUFF2 | TRFPowerDown.BUFF3 | TRFPowerDown.BUFF4
        self.__program_divider(divider_config)
        if fractional_cal is not None:
            self.__set_fractional_cal(fractional_cal)
        else:
            self.__set_fractional_cal(DEFAULT_FRACCAL)
        self.en_frac = fractional
        if calibrate:
            start_cal = time.time()
            self.en_cal = True
            while self.en_cal:
                self.r2.invalidate()
                pass
            self.tcal_total += time.time() - start_cal
        self.powerdown_parts = (TRFPowerDown.BUFF1 | TRFPowerDown.BUFF2 | TRFPowerDown.BUFF3 | TRFPowerDown.BUFF4) ^ self.outputs

    def __set_fractional_cal(self, frac_calibration: TRFFractionalCalibration):
        self.isource = True
        self.dith = True
        self.mod_ord = frac_calibration.mod_ord
        self.dith_sel = False
        self.del_sd_clk = 0b10
        self.en_frac = True
        self.ld_isource = False
        self.isource_sink = frac_calibration.isource_sink
        self.isource_trim = frac_calibration.isource_trim
        self.icpdouble = False

    @field(slice(5, 18))
    def rdiv(self):
        return self.r1

    @field_bool(19)
    def ref_inv(self):
        return self.r1

    @field_bool(20)
    def neg_vco(self):
        return self.r1

    @field(slice(21, 26))
    def icp(self):
        return self.r1

    @field_bool(26)
    def icpdouble(self):
        return self.r1

    @field(slice(27, 31))
    def cal_clk_sel(self):
        return self.r1

    @field(slice(5, 21))
    def nint(self):
        return self.r2

    @field(slice(21, 23))
    def pll_div_sel(self):
        return self.r2

    @field(slice(23, 24))
    def prsc_sel(self):
        return self.r2

    @field(slice(26, 28))
    def vco_sel(self):
        return self.r2

    @field_bool(28)
    def vco_sel_mode(self):
        return self.r2

    @field(slice(29, 31))
    def cal_acc(self):
        return self.r2

    @field_bool(31)
    def en_cal(self):
        return self.r2

    @field(slice(5, 30))
    def nfrac(self):
        return self.r3

    @field_enum(slice(5, 16), TRFPowerDown)
    def powerdown_parts(self):
        return self.r4

    @field_bool(18)
    def isource(self):
        return self.r4

    @field_bool(25)
    def dith(self):
        return self.r4

    @field(slice(26, 28))
    def mod_ord(self):
        return self.r4

    @field_bool(28)
    def dith_sel(self):
        return self.r4

    @field(slice(29, 31))
    def del_sd_clk(self):
        return self.r4

    @field_bool(31)
    def en_frac(self):
        return self.r4

    @field_bool(31)
    def ld_isource(self):
        return self.r5

    @field(slice(7, 13))
    def vco_trim(self):
        return self.r6

    @field_bool(19)
    def isource_sink(self):
        return self.r6

    @field(slice(20, 23))
    def isource_trim(self):
        return self.r6

    @field(slice(23, 25))
    def lo_div_sel(self):
        return self.r6

    @field(slice(15, 21))
    def vco_trim_readback(self):
        return self.r0

    @field(slice(22, 24))
    def vco_sel_readback(self):
        return self.r0

# import logging
# logging.basicConfig()
# logging.getLogger('mkidgen3').setLevel('DEBUG')
# import mkidgen3.ifboard
# d='/dev/cu.usbmodem14401'
# b=mkidgen3.ifboard.IFBoard(d)



import copy
from logging import getLogger
import numpy.typing as nt
import numpy as np
from mkidgen3.system_parameters import ADC_DAC_INTERFACE_WORD_LENGTH, DAC_RESOLUTION, DAC_LUT_SIZE, N_CHANNELS, \
    SYSTEM_BANDWIDTH, IF_ATTN_STEP, ADC_MAX_V, PHASE_FRACTIONAL_BITS


class Component:
    def __init__(self, name: str | None, gain: float):
        self.name = name
        self._gain = gain

    @property
    def gain(self):
        return self._gain

    def __add__(self, other):
        return Component(None, self.gain + other.gain)


class Attenuator(Component):
    def __init__(self, name: str, gain: float):
        super().__init__(name, gain)
        assert gain <= 0, "gain value should be negative for an attenutor"

    def output(self, input_power: float) -> float:
        return input_power + self.gain

    def __repr__(self):
        return f'{self.name} {self.gain} dB attn'


class ProgrammableAttenuator(Attenuator):
    def __init__(self, name: str, gain: float):
        super().__init__(name, gain)

    @property
    def gain(self):
        return self._gain

    @gain.setter
    def gain(self, gain: float):
        assert gain <= 0, "gain value should be negative for an attenutor"
        self._gain = gain


class Device(Component):
    def __init__(self, name: str, gain: float):
        super().__init__(name, gain)

    def output(self, input_power: float) -> float:
        return input_power + self.gain

    def __repr__(self):
        return f"Device ({self.gain} dB attn)"


class Amplifier(Component):
    def __init__(self, name: str, gain: float, saturation: float):
        super().__init__(name, gain)
        assert gain >= 0, "gain value should be positive for an amplifier"
        self.saturation = saturation  # dBm

    def __repr__(self):
        return f'{self.name}, gain: {self.gain} dB, saturation: {self.saturation} dBm'

    def output(self, input_power: float) -> float:
        """

        Args:
            input_power: dBm

        Returns: output power in dBm

        """
        return input_power + self.gain

    def is_saturated(self, input_power: float) -> bool:
        return self.output(input_power) > self.saturation

    @property
    def max_input(self):
        return self.saturation - self.gain


class ROChain:
    def __init__(self, *ro_chain):
        self.ro_chain = ro_chain
        self._index = 0

    @property
    def device_idx(self) -> int:
        return next(i for i, v in enumerate(self.ro_chain) if isinstance(v, Device))

    @property
    def dac_attn_idx(self) -> int:
        return next(
            i for i, v in enumerate(self.ro_chain) if (isinstance(v, ProgrammableAttenuator) & ('dac' in v.name)))

    @property
    def adc_attn1_idx(self) -> int:
        return next(
            i for i, v in enumerate(self.ro_chain) if (isinstance(v, ProgrammableAttenuator) & ('adc' in v.name)))

    @property
    def adc_attn2_idx(self) -> int:
        return len(self.ro_chain) - next(i for i, v in enumerate(self.ro_chain[::-1]) if
                                         (isinstance(v, ProgrammableAttenuator) & ('adc' in v.name))) - 1

    def __repr__(self):
        string = ''
        for component in self.ro_chain:
            string += component.__repr__() + '\n'
        return string

    def __getitem__(self, item):
        return self.ro_chain[item]

    def __iter__(self):
        return self

    def __next__(self):
        if self._index < len(self.ro_chain):
            item = self.ro_chain[self._index]
            self._index += 1
            return item
        else:
            raise StopIteration

    def validate(self, input_power):
        for component in self.ro_chain:
            try:
                if component.is_saturated(input_power):
                    raise ValueError(f'{component.name} receiving {input_power} dBm is saturated.')
            except AttributeError:
                pass
            input_power = component.output(input_power)

    def calculate_gain(self, start, stop):
        gain = 0
        for i in range(start, stop):
            gain += self.ro_chain[i].gain
        return gain

    def pre_device_gain(self):
        return self.calculate_gain(self.dac_attn_idx + 1, self.device_idx)

    def post_device_gain(self):
        return self.calculate_gain(self.device_idx, self.adc_attn1_idx)

    def post_adc_attn2_gain(self):
        return self.calculate_gain(self.adc_attn2_idx, len(self.ro_chain))


def calculate_adc_dac_attn(dac_output_power: float, device_power: float, signal_chain: ROChain,
                            adc_input_power: float) -> tuple[float, tuple[float, float]]:
    """
    Args:
        dac_output_power: dac output power [dBm]
        device_power: target power at MKID device [dBm]
        signal_chain: list of components
        adc_input_power: target adc input power [dBm]

    Returns: tuple where the first element is the total dac attenuation and the second element is a tuple where the
    first element is the adc attenuation of the first programmable attenuator in the receive chain and the second
    element is the second programmable attenuator in the receive chain.

    """
    chain = copy.deepcopy(signal_chain)
    dac_attn = -device_power + dac_output_power + chain.pre_device_gain()
    if dac_attn < 0:
        getLogger(__name__).warning(
            f'Cannot acheive device power {device_power} dBm, setting DAC attn to 0, device power will be {dac_output_power + chain.pre_device_gain()} dBm')
        dac_attn = 0
    chain[chain.dac_attn_idx].gain = -dac_attn
    post_adc_attns = adc_input_power - chain.post_adc_attn2_gain()
    pre_adc_attns = device_power + chain.post_device_gain()
    total_adc_attenuation = pre_adc_attns - post_adc_attns

    amp_between_adc_attns = chain.ro_chain[chain.adc_attn1_idx + 1]
    assert isinstance(amp_between_adc_attns, Amplifier), "adc attenuator 1 needs to be followed by amplifier"
    if pre_adc_attns > amp_between_adc_attns.max_input:
        chain.ro_chain[chain.adc_attn1_idx].gain = pre_adc_attns - amp_between_adc_attns.max_input
        chain.ro_chain[chain.adc_attn2_idx].gain = -total_adc_attenuation + chain.ro_chain[chain.adc_attn1_idx].gain
    else:
        chain.ro_chain[chain.adc_attn1_idx].gain = 0
        chain.ro_chain[chain.adc_attn2_idx].gain = -total_adc_attenuation

    chain.validate(dac_output_power)
    adc_attn = -chain.ro_chain[chain.adc_attn1_idx].gain + -chain.ro_chain[chain.adc_attn2_idx].gain  # It may make sense to return these separately

    return -chain.ro_chain[chain.dac_attn_idx].gain, adc_attn


def compute_power_sweep_attenuations(dac_attn_start: float, adc_attn_start: float,
                                     dac_attn_stop: float | None = None, points: int | None = None,
                                     step_size=IF_ATTN_STEP) -> list[tuple[float | int, float | int]]:
    """
    Args:
        dac_attn_start: starting dac attenuation [dB]
        adc_attn_start: starting adc attenuation [dB]
        dac_attn_stop: *last dac attenuation [dB]
        points: *number of attenuation steps, overrides step_size
        step_size: *attenuation step size

    Returns: list of tuples containing DAC and ADC attns
    For Ex: [(dac_attn0, adc_attn0), (dac_attn1, adc_attn1), ...]

    *A subset of these is required. Precedence is resolved like so:
    If dac_attn_stop:
        if points:
            dac_attn_stop and points determine the list and step_size is ignored
        elif step_size:
            dac_attn_stop and step_size determine the list
    else:
        step size and points determine the list


    """
    for attn in [dac_attn_start, adc_attn_start, dac_attn_stop]:
        if attn:
            assert attn >= 0, "attenuation must be positive"

    if dac_attn_stop:
        if points:
            dac_attns, step = np.linspace(dac_attn_start, dac_attn_stop, points, retstep=True)
            adc_attns = np.linspace(adc_attn_start, adc_attn_start - step * (len(dac_attns) - 1), points)
        else:
            dac_attns = np.arange(dac_attn_start, dac_attn_stop + step_size, step_size)
            adc_attns = np.arange(adc_attn_start, adc_attn_start - step_size * len(dac_attns), -step_size)
    else:
        dac_attns = np.arange(dac_attn_start, dac_attn_start + step_size * points, step_size)
        adc_attns = np.arange(adc_attn_start, adc_attn_start - step_size * points, -step_size)

    return [(dac, adc) for dac, adc in zip(dac_attns, adc_attns)]


def single_res_sweep(if_board, lo_sweep_freqs, ol):
    iq_vals = np.zeros(lo_sweep_freqs.size, dtype=np.complex64)
    for x in range(len(lo_sweep_freqs)):
        if_board.set_lo(lo_sweep_freqs[x] * 1e-6)
        iq_vals[x] = get_iq_point(ol)
    return iq_vals


def get_iq_point(ol, n=256):
    """
    Args:
        n: int
        how many points to average
    Returns: a single averaged iq data point captured from res channel 0
    """
    x = ol.capture.capture_iq(n, [0, 1], tap_location='ddciq')
    tmp = np.zeros(x.shape[1:])
    x.mean(axis=0, out=tmp)
    del x
    return tmp[0, 0] + 1j * tmp[0, 1]

def compute_lo_steps(center: float | int, resolution: float | int, bandwidth: float | int) -> nt.NDArray[float | int]:
    """
    Compute LO steps.
    Args:
        center: center frequency of the sweep bandwidth [Hz]
        resolution: frequency resolution in for the LO sweep [Hz]
        bandwidth: bandwidth for the LO to sweep through [Hz]

    Returns: list of LO settings

    """
    n_steps = np.round(bandwidth / resolution).astype('int')
    return np.linspace(-bandwidth / 2, bandwidth / 2, n_steps) + center



import logging
import pickle
import time

from logging import getLogger

from mkidgen3.server.captures import CaptureRequest
from mkidgen3.server.feedline_config import RFDCConfig
from mkidgen3.server.fpga_objects import FeedlineHardware, DEFAULT_BIT_FILE
from mkidgen3.server.misc import zpipe
from mkidgen3.util import check_active_jupyter_notebook
import asyncio
import zmq
import threading
from mkidgen3.util import setup_logging
from datetime import datetime
import argparse

COMMAND_LIST = ('reset', 'capture', 'bequiet', 'status')


class TapThread:
    def __init__(self, target, cr:CaptureRequest):
        context = zmq.Context().instance()
        a, b = zpipe(context)
        t = threading.Thread(target=target, name=f"TapThread: {cr.tap}:{cr.id}", args=(b, cr), kwargs=dict(context=context))
        self.thread = t
        self.request = cr
        self._pipe = a
        self._other_pipe = b
        t.start()

    def __repr__(self):
        a = 'running' if self.thread.is_alive() else 'stopped'
        return f'<{self.thread.name} ({a})>'

    def abort(self):
        try:
            # Abort the thread, not the request, the thread will handle the abort of the request if necessary!
            getLogger(__name__).debug(f'Sending abort to worker: {self}')
            self._pipe.send(b'abort')
        except zmq.ZMQError:
            getLogger(__name__).critical(f'Error sending abort to worker thread {self.thread}')
            raise

    def __del__(self):
        try:
            self._pipe.close()
            self._other_pipe.close()
        except zmq.error.ContextTerminated:
            pass

class FeedlineReadoutServer:
    def __init__(self, bitstream, clock_source='4.096GSPS_MTS_dualloop', if_port='dev/ifboard', ignore_version=False,
                 program_clock=True, mts=True, download=False):
        self.hardware = FeedlineHardware(bitstream, clock_source=clock_source, if_port=if_port,
                                         ignore_version=ignore_version, program_clock=program_clock,
                                         rfdc=RFDCConfig(dac_mts=mts, adc_mts=False), download=download)

        self._tap_threads = {k: None for k in ('photon', 'postage', 'engineering')}
        self._to_check = []
        self._checked = []
        self._cap_pipe=None

    def status(self):
        """

        Returns: Dictionary of status information

        """
        status = {'hardware': self.hardware.status(),
                  'running_captures': self._running_captures(),
                  'pending_captures': self._pending_captures()}

        return status

    def _running_captures(self):
        return tuple([tt.request.id for tt in list(self._tap_threads.values())])

    def _pending_captures(self):
        return tuple([cr.id for cr in self._checked+self._to_check])

    def _abort_all(self, join=False, reason='Abort all', raisezmqerror=True, also:tuple|list|CaptureRequest=tuple()):
        toabort=self._checked + self._to_check
        if isinstance(also, CaptureRequest):
            also = [also]
        toabort.extend(also)
        for cr in toabort:
            cr.establish()
            cr.abort(reason)  # signal that captures will never happen
            cr.destablish()
        self._checked, self._to_check = [], []
        # stop any running tap threads
        for tt in (tt for tt in self._tap_threads.values() if tt is not None):
            try:
                tt.abort()
            except zmq.ZMQError as e:
                if raisezmqerror:
                    raise e
        if join:
            for tt in self._tap_threads.values():
                if tt:
                    tt.thread.join()

    def _abort_by_id(self, id):
        aborted = False
        running_by_id = {tt.request.id: tt for tt in self._tap_threads.values() if tt is not None}
        if id in running_by_id:
            tt = running_by_id[id]
            aborted = True
            getLogger(__name__).debug(f'Found request {id} being serviced in {tt}. Aborting servicer.')
            tt.abort()
        for cr in filter(lambda x: x.id == id, self._checked):
            aborted = True
            getLogger(__name__).debug(f'Found request {id} in list of checked pending CR. Aborted')
            self._checked.pop(self._checked.index(cr))
            cr.abort('Abort by id')
        for cr in filter(lambda x: x.id == id, self._to_check):
            aborted = True
            getLogger(__name__).debug(f'Found request {id} in list of pending CR to be checked. Aborted')
            self._to_check.pop(self._to_check.index(cr))
            cr.abort('Abort by id')

        if not aborted:
            getLogger(__name__).info(f'Capture request {id} is unknown and cannot be aborted.')

    def create_capture_handler(self, start=True, daemon=False, context: zmq.Context = None):
        self._cap_pipe, self._cap_pipe_thread = zpipe(context or zmq.Context.instance())

        thread = threading.Thread(name='FRS CR Handler', target=self._main, args=(self._cap_pipe_thread,),
                                  kwargs={'context': context}, daemon=daemon)
        if start:
            thread.start()

        return thread

    def terminate_capture_handler(self):
        if self._cap_pipe:
            self._cap_pipe.send_pyobj(('exit', None))
            self._cap_pipe.close()

    def __del__(self):
        try:
            self._cap_pipe_thread.close()
        except:
            pass

    def abort_all(self):
        if self._cap_pipe:
            self._cap_pipe.send_pyobj(('abort', 'all'))

    def abort(self, id):
        if self._cap_pipe:
            self._cap_pipe.send_pyobj(('abort', id))

    def capture(self, capture_request):
        if self._cap_pipe:
            self._cap_pipe.send_pyobj(('capture', capture_request))

    def _cleanup_completed(self):
        """ Return true iff the effective config requirements changed """
        # Check to see if any capture threads have finished
        complete = [k for k, t in self._tap_threads.items() if t is not None and not t.thread.is_alive()]
        # for each finished capture thread remove its settings from the requirements pot and cleanup

        if bool(complete):  # need to check up to the size of the queue if anything finished
            # if the effective settings didn't change on completion we don't need to recheck stuff
            #  ignore this optimization for now
            self._to_check.extend(self._checked)
            self._checked = []

        effective_changed = False
        for k in complete:
            effective_changed |= self.hardware.derequire_config(self._tap_threads[k].request.id)
            del self._tap_threads[k]
            self._tap_threads[k] = None

        return effective_changed

    def _main(self, pipe: zmq.Socket, context: zmq.Context = None):
        """
        Enqueue a list of capture requests for future handling. Invalid requests are dealt with immediately and not
        enqueued.

        Args:
            pipe: a pipe for receiving capture requests
            context: zmq.Context

        Returns: None

        """
        context = context or zmq.Context().instance()

        #THIS LOOP SEEMS TO ACTUALLY be getting used ONLY to be deleted with the pynq.UIO reader deletion
        #when the threaded interrupt manager starts stuff up and nixes pynqs reader
        # (that apparently exists in this thread)
        try:
            aio_eloop = asyncio.get_running_loop()
        except RuntimeError:
            aio_eloop = asyncio.new_event_loop()
            getLogger(__name__).warning('Creating but not starting a thread that really should be axed and '
                                        'optimized away')
            t=threading.Thread(daemon=True, target=aio_eloop.run_forever, name='plramcap_eloop')

        asyncio.set_event_loop(aio_eloop)

        getLogger(__name__).info('Main thread starting')
        while True:

            effective_changed = self._cleanup_completed()

            running_by_id = {tt.request.id: tt for tt in self._tap_threads.values() if tt is not None}

            cr = None  # CR is the capture request that will be ckicked off this iteration of the loop
            # check for any incoming info: CapRequest, ABORT id|all, EXIT
            cmd, data = '', ''
            try:
                cmd, data = pipe.recv_pyobj(zmq.NOBLOCK)
            except zmq.ZMQError as e:
                if e.errno != zmq.EAGAIN:
                    self._abort_all(reason='Keyboard interrupt')
                    if e.errno == zmq.ETERM:
                        break
                    else:
                        raise e  # real error

            if cmd not in ('exit', 'abort', 'capture', ''):
                getLogger(__name__).error(f'Received invalid command "{cmd}"')
                cmd, data = '', ''

            if cmd == 'exit':
                self._abort_all(join=True)
                break
            elif cmd == 'abort':
                if data == 'all':
                    self._abort_all(join=True)
                else:
                    self._abort_by_id(data)
            elif cmd == 'capture':
                self.hardware.config_manager.learn(data.feedline_config)
                unknown = self.hardware.config_manager.unlearned_hashes(data.feedline_config)
                if unknown:  # We've never been sent the full config necessary
                    try:
                        data.establish()
                        #TODO this code seems to imply json "{'resp': 'ERROR', 'data': unknown}"
                        data.fail(f'ERROR: Full FeedlineConfig never sent: {unknown}')
                    except zmq.ZMQError as e:
                        getLogger(__name__).error(f'Unable to fail request with hashed config due to {e}. '
                                                  f'Silently dropping request {data.id}')
                elif (not self._to_check and self._tap_threads[data.type] is None and
                      self.hardware.config_compatible_with(data.feedline_config)):
                    cr = data  # this can be run and nothing else, so it will be done below
                else:
                    q = self._to_check if self._to_check else self._checked
                    try:
                        data.set_status('queued', f'Queued', destablish=True)
                        q.append(data)
                    except zmq.ZMQError as e:
                        getLogger(__name__).error(f'Unable to update status due to {e}. Silently dropping request'
                                                  f' {data.id}')

                # cant be run because there might be something more important (we check anyway),
                # the tap is in use (we check when the tap finishes)
                # settings aren't compatible (we will check when something finishes)

            if not cr:
                try:
                    cr = self._to_check.pop(0)
                except IndexError:
                    continue

            assert isinstance(cr, CaptureRequest)

            try:
                if self._tap_threads[cr.type] is not None:
                    cr.set_status('queued', f'tap location in use by: {self._tap_threads[cr.type].request.id}')
                    self._checked.append(cr)
                    continue
                else:
                    if not self.hardware.config_compatible_with(cr.feedline_config):
                        cr.set_status('queued', f'incompatible with one or more of: {running_by_id.keys()}')
                        self._checked.append(cr)
                        continue
            except zmq.ZMQError as e:
                getLogger(__name__).error(f'Unable to update status due to {e}. Silently aborting request {cr}.')
                continue

            cr.destablish()  # ensure nothing lingers from any status messages

            try:
                self.hardware.apply_config(cr.id, cr.feedline_config)
            except Exception as e:
                self.hardware.derequire_config(id)  # Not  necessary as we are dying, but let's die in a clean house
                getLogger(__name__).critical(f'Hardware settings failure: {e}. Aborting all requests and dying.')
                self._abort_all(reason='Hardware settings failure', raisezmqerror=False, join=False, also=cr)
                break

            self.start_tap_thread(cr)

        aio_eloop.close()
        getLogger(__name__).info('Capture thread exiting')

    def start_tap_thread(self, cr):
        """
        Start a thread to service a capture request

        Args:
            cr: a capture request.

        Returns:

        """
        assert self._tap_threads.get(cr.type, None) is None, 'Only one TapThread per location may be created at a time'
        cap_runners = {'engineering': self.hardware.plram_cap, 'photon': self.hardware.photon_cap,
                       'postage': self.hardware.postage_cap}
        target = cap_runners[cr.type]


        cr.set_status('running', f'Started at UTC {datetime.utcnow()}', destablish=True)

        self._tap_threads[cr.type] = TapThread(target, cr)


def parse_cl():
    parser = argparse.ArgumentParser(description='Feedline Readout Server', add_help=True)
    parser.add_argument('-p', '--port', dest='port', action='store', required=False, type=int,
                        help='Server port', default='8888')
    parser.add_argument('--cap_port', dest='capture_port', action='store', required=False, type=int,
                        help='Capture Data Port', default='8889')
    parser.add_argument('--sta_port', dest='status_port', action='store', required=False, type=int,
                        help='Capture Status Port', default='8890')
    parser.add_argument('--clock', dest='clock', action='store', required=False, type=str,
                        help='Clock Source', default='default')
    parser.add_argument('-b', '--bitstream', dest='bitstream', action='store', required=False, type=str,
                        help='bitstream file',
                        default=DEFAULT_BIT_FILE)
    parser.add_argument('--if', dest='ifboard', action='store', required=False, type=str,
                        help='IF Board device', default='/dev/ifboard')
    parser.add_argument('--iv', dest='ignore_fpga_driver_version', action='store_true', required=False,
                        help='Ignore FPGA driver version checks', default=False)
    return parser.parse_args()


def start_zmq_devices(cap_addr, stat_addr):
    from zmq.devices import ThreadDevice

    cap_addr_internal = 'inproc://cap_data.xsub'
    stat_addr_internal = 'inproc://cap_stat.xsub'
    # Set up a proxy for routing all the capture requests
    dtd = ThreadDevice(zmq.QUEUE, zmq.XSUB, zmq.XPUB)
    dtd.setsockopt_in(zmq.LINGER, 0)
    dtd.setsockopt_out(zmq.LINGER, 0)
    dtd.bind_in(cap_addr_internal)
    dtd.bind_out(cap_addr)
    dtd.daemon = True
    dtd.start()
    getLogger(__name__).info(f'Publishing capture data to {cap_addr} from relay @ {cap_addr_internal}')

    std = ThreadDevice(zmq.QUEUE, zmq.XSUB, zmq.XPUB)
    std.setsockopt_in(zmq.LINGER, 0)
    std.setsockopt_out(zmq.LINGER, 0)
    std.bind_in(stat_addr_internal)
    std.bind_out(stat_addr)
    std.daemon = True
    std.start()
    getLogger(__name__).info(f'Publishing capture status information to {stat_addr} from relay @ {stat_addr_internal}')

    return dtd, std


if __name__ == '__main__':

    import os, time
    os.environ['TZ'] = 'right/UTC'
    time.tzset()
    setup_logging('feedlinereadoutserver')

    if check_active_jupyter_notebook():
        raise RuntimeError('Jupyter notebooks are running, shut them down first.')

    args = parse_cl()

    context = zmq.Context.instance(io_threads=2)
    context.linger = 1

    fr = FeedlineReadoutServer(args.bitstream, clock_source=args.clock, if_port=args.ifboard,
                               ignore_version=args.ignore_fpga_driver_version, program_clock=True, mts=True,
                               download=False)

    # Set up proxies for routing all the capture data and status
    cap_addr = f'tcp://*:{args.capture_port}'
    stat_addr = f'tcp://*:{args.status_port}'
    start_zmq_devices(cap_addr, stat_addr)

    # Set up a command port
    command_port = args.port
    cmd_addr = f"tcp://*:{command_port}"
    socket = context.socket(zmq.REP)
    socket.bind(cmd_addr)

    # Start up the main thread
    thread = fr.create_capture_handler(context=context, start=True, daemon=False)

    getLogger(__name__).info(f'Accepting commands on {cmd_addr}')

    while True:
        try:
            cmd, arg = socket.recv_pyobj()
        except zmq.ZMQError as e:
            getLogger(__name__).error(f'Caught {e}, aborting and shutting down')
            fr.terminate_capture_handler()
            break
        except KeyboardInterrupt:
            getLogger(__name__).error(f'Keyboard Interrupt, aborting and shutting down')
            fr.terminate_capture_handler()
            break
        except pickle.UnpicklingError:
            socket.send_pyobj('ERROR: Ignoring unpicklable command')
            getLogger(__name__).error(f'Ignoring unpicklable command')
            continue
        else:
            if not thread.is_alive():
                getLogger(__name__).critical(f'Capture thread has died prematurely. All existing captures will '
                                             f'never complete. Exiting.')
                socket.send_pyobj('ERROR')
                break

        getLogger(__name__).debug(f'Received command "{cmd}" with args {arg}')

        if cmd == 'reset':
            socket.send_pyobj('OK')
            fr.terminate_capture_handler()
            thread.join()
            fr.hardware.reset()
            thread = fr.create_capture_handler(context=zmq.Context.instance(), start=True, daemon=False)

        elif cmd == 'status':
            try:
                status = fr.status()  # this might take a while and fail
            except Exception as e:
                status = {'hardware': str(e)}
            status['id'] = f'FRS {args.fl_id} @ {args.port}/{args.cap_port}'
            socket.send_pyobj(status)

        elif cmd == 'bequiet':
            fr.abort_all()
            try:
                fr.hardware.bequiet(**arg)  # This might take a while and fail
                socket.send_pyobj('OK')
            except Exception as e:
                socket.send_pyobj(f'ERROR: {e}')

        elif cmd == 'capture':
            fr.capture(arg)
            socket.send_pyobj({'resp': 'OK', 'code': 0})

        elif cmd == 'abort':
            fr.abort(arg)
            socket.send_pyobj({'resp': 'OK', 'code': 0})

        else:
            socket.send_pyobj({'resp': 'ERROR', 'code': 0})

    thread.join()
    socket.close()
    context.term()



import asyncio
from logging import getLogger
import numpy as np
import pynq

from mkidgen3.util import _which_one_bit_set
from pynq import allocate, DefaultIP
import bitstruct
from pynq.lib.dma import _SGDMAChannel

class MCMM2SBufferDescriptor:
    """
    next pointer 32bits
        0:5 Must be 0, descriptors must be 16-word aligned that is, 0x00, 0x40, 0x80, and so forth. Any other
            alignment has undefined results.
        6:31 value must be 16word aligned (e.g multiple of 0x40)
    next pointer upper 32
    buffer addr 32bits
        - must be data width aligned if core lacks Buffer Realignment Engine (ours do)
    buffer addr upper 32
    reserved 32bits set to 0
    control 32bits
        0:25 Buffer Length Indicates the size in bytes of the transfer buffer. This value indicates the amount of
            bytes to transmit out on the MM2S stream. The usable width of buffer length is specified by the
            parameter Width of Buffer Length Register in the block design. A maximum of 67,108,863 bytes can be
            described by this field. This value should be an integer multiple of the AXI4-Stream data width;
            however it can have any value if the TX EOF bit is set.
        26:29 Reserved
        30 TX EoF End of Frame. Flag indicating the last buffer to be processed. Set this flag to indicate to AXI
            MCDMA that this descriptor describes the end of the packet. The buffer associated with this descriptor is
            transmitted last and results in TLAST assertion.
        31 TX SoF Start of Frame. Flag indicating the first buffer to be processed. Set this flag to indicate to the
            AXI MCDMA that this descriptor describes the start of the packet. The buffer associated with this descriptor
            is transmitted first.
    control sideband 32bits
        0:15 TUSER This field contains the value of TUSER to be presented on the last data beat of the packet that is,
            the beat that has TLAST.
        16:23 Reserved.
        24:31 TID This field contains the value of TID to be presented on the AXI4-Stream interface. The value of TID
            remains the same throughout the packet length. Care should be taken to ensure that this value remains
            constant in the descriptor chain for a packet (e.g. between TLAST).
    status 32bits (set by core)
        0:25 Transferred Bytes Indicates the size in bytes of the actual data transferred for this descriptor. This
            value indicates the amount of bytes to transmit out on MM2S stream. This value should match the Control
            Buffer Length field. The usable width of Transferred Bytes is specified by the parameter Width of Buffer.
        26:27 Reserved
        28 DMAIntErr MCDMA Internal Error. Internal Error detected by AXI DataMover. This error can occur if a 0
            length bytes to transfer is fed to the AXI DataMover. This only happens if the Buffer Length specified in
            the fetched descriptor is set to 0. This error can also be caused if there is an under-run or
            over-run condition. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is
            set to 0, and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1.
        29 DMASlvErr DMA Slave Error. Slave Error detected by primary AXI DataMover. This error occurs if the slave
            read from the Memory Map interface issues a Slave Error. This error condition causes the AXI MCDMA to
            halt gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down, the
            MCDMASR.Halted bit is set to 1.
        30 DMADecErr DMA Decode Error. Decode Error detected by primary AXI DataMover. This error occurs if the
            Descriptor Buffer Address points to an invalid address. This error condition causes the AXI MCDMA to halt
            gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down,
            the MCDMASR.Halted bit is set to 1.
        31 Complete This indicates that the MCDMA engine has completed the transfer as described by the
            associated descriptor. The MCDMA Engine sets this bit to 1 when the transfer is completed. The software can
            manipulate any descriptor with the Completed bit set to 1. If a descriptor is fetched with this bit set to
            1, the descriptor is considered a stale descriptor. A SGIntErr is flagged and the AXI MCDMA engine halts.
    User0 - User4 32bits each
        User application fields 0 to 4. Specifies user-specific application data. When Status Control Stream is
        enabled, the Application (APP) fields of the Start of Frame (SOF) Descriptor are transmitted to the AXI
        Control Stream. For other MM2S descriptors with SOF = 0, the APP fields are fetched but ignored. See PG288 for
        more information.
    """
    pass


class MCS2MMBufferDescriptor:
    """
    S2MM BD:
    next pointer 32bits
        0:5 Must be 0, descriptors must be 16-word aligned that is, 0x00, 0x40, 0x80, and so forth. Any other
            alignment has undefined results.
        6:31 value must be 16word aligned (e.g multiple of 0x40)
    next pointer upper 32
    buffer addr 32bits
        - must be data width aligned if core lacks Buffer Realignment Engine (ours do)
    buffer addr upper 32
    reserved 32bits set to 0
    control 32bits
        0-25 This value indicates the amount of space in bytes available for receiving data in an S2MM stream.
            The usable width of buffer length is specified by the parameter Width of Buffer Length Register
            (c_sg_length_width). A maximum of 67,108,863 bytes of transfer can be described by this field. This
            value should be an integer multiple of AXI4-Stream data width. Note: The total buffer space in the S2MM
            descriptor chain (that is, the sum of buffer length values for each descriptor in a chain) must be, at a
            minimum, capable of holding the maximum receive packet size. Undefined results occur if a packet larger
            than the defined buffer space is received. Setting the Buffer Length Register Width smaller than 26
            reduces FPGA resource utilization.
        26:31 Reserved
    status 32bits (set by core)
        0:25 Transferred Bytes This value indicates the amount of data received and stored in the buffer described by
            this descriptor. This might or might not match the buffer length. For example, if this descriptor
            indicates a buffer length of 1,024 bytes but only 50 bytes were received and stored in the buffer,
            then the Transferred Bytes field indicates 0x32. The entire receive packet length can be determined by
            adding the Transferred Byte values from each descriptor from the RXSOF descriptor to the Receive End of
            Frame (RXEOF) descriptor. The usable width of Transferred Bytes is specified by the parameter
            Width of Buffer Length Register (c_sg_length_width) in the block design. A maximum of 67,108,863 bytes of
            transfer can be described by this field.
        26 RXEOF End of Frame. Flag indicating buffer holds the last part of packet. This bit is set by AXI MCDMA to
            indicate that the buffer associated with this descriptor contains the end of the packet. User Application
            data sent through the status stream input is stored in APP0 to APP4 of the RXEOF descriptor when the
            Control/Status Stream is enabled.
        27 RXSOF Start of Frame. Flag indicating buffer holds first part of packet. This bit is set by AXI MCDMA to
            indicate that the buffer associated with this descriptor contains the start of the packet.
        28 DMAIntErr MCDMA Internal Error. Internal Error detected by AXI DataMover. This error can occur if a 0
            length bytes to transfer is fed to the AXI DataMover. This only happens if the Buffer Length specified in
            the fetched descriptor is set to 0. This error can also be caused if there is an under-run or
            over-run condition. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is
            set to 0, and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1.
        29 DMASlvErr DMA Slave Error. Slave Error detected by primary AXI DataMover. This error occurs if the slave
            read from the Memory Map interface issues a Slave Error. This error condition causes the AXI MCDMA to
            halt gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down, the
            MCDMASR.Halted bit is set to 1.
        30 DMADecErr DMA Decode Error. Decode Error detected by primary AXI DataMover. This error occurs if the
            Descriptor Buffer Address points to an invalid address. This error condition causes the AXI MCDMA to halt
            gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down,
            the MCDMASR.Halted bit is set to 1.
        31 Complete This indicates that the MCDMA engine has completed the transfer as described by the
            associated descriptor. The MCDMA Engine sets this bit to 1 when the transfer is completed. The software can
            manipulate any descriptor with the Completed bit set to 1. If a descriptor is fetched with this bit set to
            1, the descriptor is considered a stale descriptor. A SGIntErr is flagged and the AXI MCDMA engine halts.
    sideband 32bits (set by core)
        0:15 TUSER This field contains the value of TUSER present on the last data beat of the packet that is,
            the beat that has TLAST.
        16:19 TDEST This field contains the value of TDEST.
        20:23 Reserved.
        24:31 TID This field contains the value of TID present on the AXI4-Stream input. It is expected that this
            value should remain constant throughout the packet length.
    User0 - User3 32bits each
        When Status/Control Stream is enabled, the status data received on the AXI Status Stream is stored into the
        APP fields of the End of Frame (EOF) Descriptor. For other S2MM descriptors with EOF = 0, the APP fields are
        set to zero by the Scatter Gather Engine.
    User4 32bits
        User Application field 4 and Receive Byte Length. If Use RxLength In Status Stream is not enabled, this field
        functions identically to APP0 to APP3 in that the status data received on the AXI Status Stream is stored into
        the APP4 field of the End of Frame (EOF) Descriptor. This field has a dual purpose when Use RxLength in
        Status Stream is enabled. The first least significant bits specified in the Buffer Length Register Width
        (up to a maximum of 16) specify the total number of receive bytes for a packet that were received on the
        S2MM primary data stream. Second, the remaining most significant bits are User Application data.
    """
    bd = bitstruct.compile('>u64 u64 p32'
                           'p6 u26'
                           'b1b1 b1b1 b1b1 u26'
                           'u8 p4 u4 u16'
                           'u32 u32 u32 u32 u32', names=['next', 'buff', 'control.buff_len',
                                                         'status.complete', 'status.dma_decode_err',
                                                         'status.dma_slave_err', 'status.dma_internal_err',
                                                         'status.rx_sof', 'status.rx_eof',
                                                         'status.xfer_len',
                                                         'sideband.tid', 'sideband.tdest', 'sideband.tuser', 'app0',
                                                         'app1', 'app2', 'app3', 'app4'])

    def __init__(self, next_addr, buf_addr, length):
        """
        Note that this may generate an illegal BD:
            next must be in multiples of 0x40
            Buf_addr mod stream_width_bytes must be 0. E.g. for a 64 byte wide stream it must be in multiples of 0x40
            length must be representable by the width of the core's Buffer Length Register
        """
        # NB Changing the hard coded defaults will set non-zero initial values for BD properties that are set by the
        # MCDMA core, while this probably has no effect it may make it harder to debug what comes back.
        # setting status.complete=True would initialize the BD as stale.
        byt = self.bd.pack({'next': next_addr, 'buff': buf_addr, 'control.buff_len': length,
                            'status.xfer_len': 0, 'status.rx_eof': False, 'status.rx_sof': False,
                            'status.dma_internal_err': False, 'status.dma_slave_err': False,
                            'status.dma_decode_err': False, 'status.complete': False, 'sideband.tuser': 0,
                            'sideband.tdest': 0, 'sideband.tid': 0, 'app0': 0, 'app1': 0, 'app2': 0, 'app3': 0,
                            'app4': 0})
        self._bytes = byt

    @staticmethod
    def from_dict(d):
        ret = MCS2MMBufferDescriptor(0, 0, 0)
        ret._bytes = ret.bd.pack(d)
        return ret

    @staticmethod
    def array_to_dict(array):
        return {'next': np.uint64(array[0]) | (np.uint64(array[1]) << np.uint64(32)),
                'buff': np.uint64(array[2]) | (np.uint64(array[3]) << np.uint64(32)),
                'control.buff_len': array[5] & ((1 << 26) - 1),
                'status.xfer_len': array[6] & ((1 << 26) - 1),
                'status.rx_eof': bool(array[6] & (1 << 26)), 'status.rx_sof': bool(array[6] & (1 << 27)),
                'status.dma_internal_err': bool(array[6] & (1 << 28)),
                'status.dma_slave_err': bool(array[6] & (1 << 29)),
                'status.dma_decode_err': bool(array[6] & (1 << 30)), 'status.complete': bool(array[6] & (1 << 31)),
                'sideband.tuser': array[7] & 0xffff, 'sideband.tdest': (array[7] >> 16) & 0xf,
                'sideband.tid': (array[7] >> 24), 'app0': array[8],
                'app1': array[9], 'app2': array[10], 'app3': array[11], 'app4': array[12]}

    @staticmethod
    def from_array(array):
        if array.size < 13:
            raise ValueError('Insufficient data, 13 words required')
        ret = MCS2MMBufferDescriptor(0, 0, 0)
        ret._bytes = ret.bd.pack(ret.array_to_dict(array))
        return ret

    @property
    def array(self):
        """Return 16 words with the contents of the buffer descriptor."""
        words = [int(self._bytes[i * 4:(i + 1) * 4].hex(), 16) for i in range(13)] + [0] * 3
        words[:2] = words[:2][::-1]  # gotta swap so that the high word goes into the right register
        words[2:4] = words[2:4][::-1]
        return np.array(words, dtype=np.uint32)


class MCS2MMBufferChain:
    def __init__(self, n, buffer_size=8192, contiguous=True, zero_buffers=True, mmio=None):
        if buffer_size % 0x40:
            raise ValueError('Buffer size must be a multiple of the stream width (0x40)')
        self._chain = allocate(16 * n, dtype=np.uint32)  # TODO ensure starting address is on a 0x40 boundary
        if self._chain.device_address % 0x40:
            raise ValueError('Chain does not start on a 0x40 address boundary')
        # if contiguous:
        #     self.contiguous = True
        #     if mmio is None:
        #         # TODO ensure starting address is on a stream width (here 0x40) boundary as DRE is disabled
        #         # TODO allocate in the PL DDR4
        #         raise NotImplementedError
        #         buff = allocate(buffer_size * n, np.uint8)
        #         self._buffers = [buff]
        #         self.bd_addr = [buff.device_address + buffer_size * i for i in range(n)]
        #     else:
        self.contiguous = True
        self._buffers = [mmio.array]
        self.bd_addr = [mmio.base_addr + buffer_size * i for i in range(n)]
        self.buffer_sizes = [buffer_size]*n
        # else:
        #     # TODO ensure starting address is on a stream width (here 0x40) boundary as DRE is disabled
        #     # TODO allocate in the PL DDR4
        #     raise NotImplementedError
        #     self._buffers = [allocate(buffer_size, np.uint8) for i in range(n)]
        #     self.bd_addr = [b.device_address for b in self._buffers]
        #     self.contiguous = False

        # Zero the buffers
        if zero_buffers:
            for b in self._buffers:
                b[:] = 0

        # Build chain
        for i, buf_addr in zip(range(n), self.bd_addr):
            descriptor = MCS2MMBufferDescriptor(self._chain.device_address + 0x40 * (i + 1), buf_addr, buffer_size)
            self._chain[i * 16:(i + 1) * 16] = descriptor.array

    @property
    def chain_length(self):
        """How many links are in the chain"""
        return len(self.bd_addr)

    @property
    def head_addr(self):
        """ Return the memory address of the first descriptor in the chain"""
        return self._chain.device_address

    @property
    def tail_addr(self):
        """ Return the memory address of the last descriptor in the chain"""
        return self._chain.device_address+16*(self.chain_length-1)*4

    def descriptor_dict(self, i):
        """Return a dict of link i of the chain"""
        if not 0 <= i < len(self.bd_addr):
            raise ValueError(f'Chain is of length {len(self.bd_addr)}')
        return MCS2MMBufferDescriptor.array_to_dict(self._chain[16*i:16*(i+1)])

    def descriptor_data(self, i):
        """Retrieve the data from the descriptor"""
        if not 0 <= i < len(self.bd_addr):
            raise ValueError(f'Chain is of length {len(self.bd_addr)}')
        d=self.descriptor_dict(i)[''] # TODO access directly for speed?
        if not d['complete']:
            raise ValueError('Descriptor not complete')
        n = d['status.xfer_len']
        return [self._buffers[0][i] for i in range(n // 4)]

    def __del__(self):
        for b in self._buffers:
            b.freebuffer()


class MCDMA(DefaultIP):
    bindto = ['xilinx.com:ip:axi_mcdma:1.1']
    ADDR_S2MM_CONTROL = 0x500
    ADDR_S2MM_COMMON_STATUS = 0x504
    ADDR_S2MM_CHAN_ENABLE = 0x508
    ADDR_S2MM_INPROGRESS = 0x50C
    ADDR_S2MM_ERROR = 0x510
    ADDR_S2MM_ALLPACKETDROP = 0x514
    ADDR_S2MM_CHAN_COMPLETED = 0x518
    ADDR_S2MM_INTMON = 0x520

    ADDR_S2MM_CHAN_CONTROL = {i: 0x540 + i * 0x40 for i in range(16)}
    ADDR_S2MM_CHAN_STATUS = {i: 0x544 + i * 0x40 for i in range(16)}
    ADDR_S2MM_CURDESC = {i: 0x548 + i * 0x40 for i in range(16)}
    ADDR_S2MM_TAILDESC = {i: 0x550 + i * 0x40 for i in range(16)}
    ADDR_S2MM_PACKETDROP = {i: 0x558 + i * 0x40 for i in range(16)}
    ADDR_S2MM_PACKETPROC = {i: 0x55C + i * 0x40 for i in range(16)}

    S2MM_ERROR_REG = bitstruct.compile('>p25 b1b1b1 p1 b1b1b1',
                                       names=['sg_decode', 'sg_slave', 'sg_internal', 'dma_decode', 'dma_slave',
                                              'dma_internal'])
    S2MM_CHAN_CTRL_REG = bitstruct.compile('>u8 u8 u8 b1b1 b1b1b1 p2 b1',
                                           names=['delay_timeout', 'bdcomplete_threshold', 'drop_threshold', 'error',
                                                  'timeout', 'bd_complete', 'packet_drop', 'otherchan_error',
                                                  'runstop'])
    S2MM_CHAN_STAT_REG = bitstruct.compile('>u8 u8 u8 b1b1b1b1 b1p1b1b1',
                                           names=['delay_timeout', 'bdcomplete_threshold', 'drop_threshold',
                                                  'error', 'timeout', 'bd_complete', 'packet_drop',
                                                  'otherchan_error', 'bd_shortfall', 'idle'])
    S2MM_CHAN_STAT_REG = bitstruct.compile('>u8 u8 u8 b1b1b1b1 b1p1b1b1',
                                           names=['delay_timeout', 'bdcomplete_threshold', 'drop_threshold',
                                                  'error', 'timeout', 'bd_complete', 'packet_drop',
                                                  'otherchan_error', 'bd_shortfall', 'idle'])

    S2MM_INT_SOURCE = bitstruct.compile('>p16'+'b1'*16, names=[f'ch{i}' for i in range(16)])
    S2MM_CHAN_COMPLETE = bitstruct.compile('>p16'+'b1'*16, names=[f'ch{i}' for i in range(16)])
    S2MM_STAT_REG = bitstruct.compile('>p30b1b1', names=['idle', 'halted'])

    def s2mm_current_descriptor(self, chan):
        return self.read(MCDMA.ADDR_S2MM_CURDESC[chan]) | (self.read(MCDMA.ADDR_S2MM_CURDESC[chan]+4)<<32)

    def s2mm_status(self):
        """
        32:2 Reserved
        1 Idle  MCDMA S2MM Idle. Indicates the state of AXI MCDMA operations. When IDLE, indicates the SG Engine has
        reached the tail pointer for all the channels and all queued descriptors have been processed.
        Writing to the tail pointer register to any channel automatically restarts MCDMA operations.
        0 Halted (MCDMA.Halted)  MCDMA Halted. Indicates the run/stop state of the MCDMA.
        There can be a lag of time between when MCDMACR.RS = 0 and when MCDMASR.Halted = 1.
        """
        return MCDMA.S2MM_STAT_REG.unpack(self.read(MCDMA.ADDR_S2MM_COMMON_STATUS).to_bytes(4,'big'))

    def s2mm_channel_status(self, chan):
        """
        31:24 IRQ DELAY Status Interrupt delay time Status. Indicates current interrupt delay time value.
        23:16 IRQ Threshold Status Interrupt Threshold Status. Indicates current interrupt threshold value.
        15:8 IRQ Packet Drop Status Interrupt Packet Drop Threshold. Indicates current interrupt threshold value.
        7 Err Irq Interrupt on Error. When set to 1, indicates an interrupt event was generated on an error.
        If the corresponding bit in Control register is enabled (Err_IrqEn = 1), an interrupt out is
        generated from the AXI MCDMA. Writing a 1 to this bit clears it.
        6 DlyIrq Interrupt on Delay. When set to 1, indicates an interrupt event was generated on delay timer
        timeout. If the corresponding bit in Control register is enabled (Dly_IrqEn = 1), an interrupt out is
        generated from the AXI MCDMA. Writing a 1 to this bit clears it.
        5 IOC_Irq Interrupt on Complete. When set to 1 an interrupt event was generated on completion of a
        descriptor. This occurs for descriptors with the EOF bit set. If the corresponding bit in the Control
        register is enabled (IOC_IrqEn = 1) and if the interrupt threshold has been met, causes an interrupt out to
        be generated from the AXI MCDMA. Writing a 1 to this bit clears it.
        4 Pktdrop_irq Interrupt on packet drop. When set to 1 indicates an interrupt event was generated on packet
        drop. If the corresponding bit in the Control register is enabled (PktDrp_IrqEn = 1) and if the interrupt
        threshold has been met, causes an interrupt out to be generated from the AXI MCDMA. Writing a 1 to this bit
        clears it.
        3 Err_on_other_ch_irq Interrupt on Error on other channels. When set to 1, indicates an interrupt event was
        generated on an error on other channels. If the corresponding bit in Control register is enabled (
        Err_on_other En = 1), an interrupt out is generated from the AXI MCDMA. Writing a 1 to this bit clears it.
        2 Reserved
        1 BD ShortFall This bit is set when a packet is being processed and the BD queue becomes empty. This means
        that the packet that is being serviced is too large to be accommodated in the BD queue. This scenario leads
        to MCDMA waiting forever for the packet to get completed. To get over this, extend the BD chain and program
        the TD to fetch more BDs. This does not result in a packet drop. This bit can also get set momentarily when
        the MCDMA is servicing the last BD for a channel and accommodates the packet. After the TLAST is
        accommodated in the last BD, this bit is unset.
        0 Idle (Queue Empty) MCDMA Channel Idle. Indicates the SG Engine has reached the tail pointer for the
        associated channel and all queued descriptors have been processed. This means that the BD queue is empty and
        there are no more BDs to process. Writing to the tail pointer register automatically restarts the BD fetch
        operations. If the packet arrives while this bit is set, that packet is dropped.
        """
        return MCDMA.S2MM_CHAN_STAT_REG.unpack(self.read(MCDMA.ADDR_S2MM_CHAN_STATUS[chan]).to_bytes(4, 'big'))

    def s2mm_packets_processed(self, chan):
        """
        31:16 RSVD
        15:0 Packet processed count Reports the number of packets processed for the channel. This counter rolls over \
                to 0 after reaching maximum value.
        """
        return self.read(MCDMA.ADDR_S2MM_PACKETPROC[chan]) & 0xffff

    def s2mm_packets_dropped(self, chan):
        """
        31:16 RSVD
        15:0 Packet drop count Reports the number of packets dropped for the channel. This value increments by 1
        every time a packet is dropped. The counter wraps around after it has reached maximum value. This register is cleared when read.
        """
        return self.read(MCDMA.ADDR_S2MM_PACKETDROP[chan]) & 0xffff

    def s2mm_total_packets_dropped(self):
        """
        Reports th number of packets dropped across all channels. This value increments by 1 every time a packet is
        dropped on any channel. The counter wraps around after it has reached maximum value.
        """
        return self.read(MCDMA.ADDR_S2MM_ALLPACKETDROP)

    def s2mm_inprogress(self):
        """
        31:16 Reserved
        15:0 This is the channel ID that was last serviced. This register has a one-hot value to identify the channel
        that caused the error. A value of 1 corresponds to TDEST = 0, a value of 2 corresponds to TDEST = 1,
        a value of 4 corresponds to TDEST = 2 and so on.
        """
        return _which_one_bit_set(self.read(MCDMA.ADDR_S2MM_INPROGRESS) & 0xffff, 16)

    def s2mm_interrupt_source(self):
        """
        This gives info about the channel(s) that generated the interrupt. This register has a one-hot value to
        identify the channel(s) that generated the interrupt. A value of 1 corresponds to Channel 0, a value of 2
        corresponds to Channel 1, a value of 4 corresponds to Channel 2 and so on. Bits are automatically cleared
        when the corresponding interrupt is cleared in the channel status register.
        """
        #return _which_one_bit_set(self.read(MCDMA.ADDR_S2MM_INTMON) & 0xffff, 16)
        return MCDMA.S2MM_CHAN_COMPLETE.unpack(self.read(MCDMA.ADDR_S2MM_INTMON).to_bytes(4, 'big'))

    def s2mm_channels_completed(self):
        """
        This is the channel that was serviced. This register has a one-hot value to identify the channel that was
        processed. A value of 1 corresponds to TDEST = 0, a value of 2 corresponds to TDEST = 1, a value of 4
        corresponds to TDEST = 2 and so on. This register is cleared on read.
        """
        # return _which_one_bit_set(self.read(MCDMA.ADDR_S2MM_CHAN_COMPLETED) & 0xffff, 16)
        return MCDMA.S2MM_CHAN_COMPLETE.unpack(self.read(MCDMA.ADDR_S2MM_CHAN_COMPLETED).to_bytes(4,'big'))

    def s2mm_error(self):
        """
        31:7 Reserved
        6 SGDecErr Scatter Gather Decode Error. This error occurs if CURDESC_PTR and/or NXTDESC_PTR point to an
        invalid address. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is set to
        0 and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1. MCDMA Engine halts. This
        error cannot be logged into the descriptor.
        5 SGSlvErr Scatter Gather Slave Error. This error occurs if the slave read from on the Memory Map interface
        issues a Slave Error. This error condition causes the AXI MCDMA to gracefully halt. The MCDMACR.RS bit is
        set to 0, and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1. MCDMA Engine
        halts.This error cannot be logged into the descriptor.
        4 SGIntErr Scatter Gather Internal Error. This error occurs if a descriptor with the Complete bit already
        set is fetched. This indicates to the SG Engine that the descriptor is a tail descriptor. This error
        condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is set to 0, and when the engine has
        completely shut down, the MCDMASR.Halted bit is set to 1. This error cannot be logged into the descriptor.
        3 Reserved
        2 DMA Dec Err MCDMA Decode Error. This error occurs if the address request points to an invalid address.
        This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is set to 0, and when the
        engine has completely shut down, the MCDMASR.Halted bit is set to 1. This bit can be set when such an event
        occurs on any of the channels.
        1 DMA SLv Err MCDMA Slave Error. This error occurs if the slave read from the Memory Map interface issues a
        Slave Error. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is set to 0
        and when the engine has completely shut down the MCDMASR.Halted bit is set to 1. This bit can be set when
        such an event occurs on any of the channels.
        0 DMA Intr Err MCDMA Internal Error. This error occurs if the buffer length specified in the fetched
        descriptor is set to 0. Also, when in Scatter Gather Mode and using the status app length field, this error
        occurs when the Status AXI4-Stream packet RxLength field does not match the S2MM packet being received by
        the S_AXIS_S2MM interface. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit
        is set to 0, and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1. This bit is
        set when such an event occurs on any of the channels.
        """
        return MCDMA.S2MM_ERROR_REG.unpack(self.read(MCDMA.ADDR_S2MM_ERROR).to_bytes(4, 'big'))

    @staticmethod
    def gen_s2mm_control_reg(runstop=False, other_chan_errors=False, packet_drop=False, bd_complete=True, timeout=False,
                             error=True, drop_threshold=1, bdcomplete_threshold=1, delay_timeout_us=100,
                             MCDMA_SGCLOCK_MHZ=128):
        """
        other_chan_errors - IRQ on other channel errors
        packet_drop = IRQ on packet drops
        bd_complete - IRQ on completion of buffer descriptors
        timeout - IRQ on no inbound packet timeout (i.e. time from tlast of one to first of next)
        error - IRQ on errors

        0 CHANNEL.RunStop (CH.RS)
        1:2 Reserved
        3 Err_on_other En
        4 Pktdrop_IrqEn
        5 IOC_IrqEn
        6 DlyIrqEn
        7 Err Irq En
        15:8 IRQ Packet Drop Threshold
        23:16 IRQ Threshold
        31:24  IRQ DELAY
        """
        delay_timeout = (delay_timeout_us * MCDMA_SGCLOCK_MHZ) // 125
        if delay_timeout == 0 and timeout:
            getLogger(__name__).warning('Channel delay timeout is quantized in units of 1.25 us (assuming a 100MHz '
                                        'SG clock. Setting to minimum. If disable intended set irq_delay=False')
        if delay_timeout > 255 and timeout:
            getLogger(__name__).warning('Maximum channel delay timeout is 318.75 us (assuming a 100MHz '
                                        'SG clock). Setting to maximum.')
            delay_timeout = 255

        drop_threshold = min(drop_threshold, 255)
        bdcomplete_threshold = min(bdcomplete_threshold, 255)

        reg = MCDMA.S2MM_CHAN_CTRL_REG.pack({'delay_timeout': delay_timeout,
                                             'bdcomplete_threshold': bdcomplete_threshold,
                                             'drop_threshold': drop_threshold, 'error': error, 'timeout': timeout,
                                             'bd_complete': bd_complete, 'packet_drop': packet_drop,
                                             'otherchan_error': other_chan_errors, 'runstop': runstop})
        return reg

    def __init__(self, description):
        super().__init__(description=description)
        self.bd_chains = None

    def config_receive(self, n_buffers=2, buffer_size_bytes=8192, channels=tuple(range(16))):
        """
        1. Enable the required channels. (can be also done after step 6).
        2. Program the CD registers of the channels. If the IP is configured for address_width > 32 (c_addr_width > 32),
            then program the corresponding MSB registers.
        3. Program the CHANNEL.RS bit of channel control registers. Note: At this point the CDs cannot be re-programmed.
        4. Start the MCDMA by programming MCDMA.RS bit (0x500).
        5. Program the interrupt thresholds, Enable Interrupts.
        6. Program the TD register of channels. If the IP is configured for address_width > 32 (c_addr_width > 32), then
           program the corresponding MSB registers. Programming the TDs of a particular channel triggers the fetching
            of the BDs for the respective channels.
        """
        # build buffer descriptor chains
        self.bd_chains = {c: MCS2MMBufferChain(n=n_buffers, buffer_size=buffer_size_bytes, contiguous=True)
                          for c in channels}

        # Enable channels
        # If the MCDMA receives a packet on a disabled channel, the entire packet is dropped by the MCDMA engine.
        # This register does not stop the BD fetch of the channel; it only stops a particular channel from being
        # serviced. This register can be programmed at any time, but it will come into effect only when there is no
        # data present on the S2MM AXI4-Stream interface or on arrival of the next packet.
        # Note: Disabling a channel does not disable its interrupt behavior. Setting a bit to 1 enables the channel
        # for service. Each bit corresponds to a channel. Bit[0] corresponds to channel with TDEST = 0 and so on.

        # 1. Enable the required channels. (can be also done after step 6). If done after step 6 BD fetching will start
        # and the core will be operating however it will drop any inbound packets until the channels are enabled.
        enable = 0
        for c in channels:
            enable |= 1 << c
        self.write(MCDMA.ADDR_S2MM_CHAN_ENABLE, enable)

        # 2. Program the CD registers of the channels.
        for c in channels:
            self.write(MCDMA.ADDR_S2MM_CURDESC[c], self.bd_chains[c].head_addr & 0xffffffff)
            self.write(MCDMA.ADDR_S2MM_CURDESC[c]+4, self.bd_chains[c].head_addr >> 32)

        # 3. Program the CHANNEL.RS bit of channel control registers. After this point the CDs cannot be reprogrammed.
        for c in range(16):
            x = self.read(MCDMA.ADDR_S2MM_CHAN_CONTROL[c])
            x = x | 1 if c in channels else (x | 1) ^ 1  # clear run bits on other channels
            self.write(MCDMA.ADDR_S2MM_CHAN_CONTROL[c], x)

        # 4. Start the MCDMA by programming MCDMA.RS bit.
        self.write(MCDMA.ADDR_S2MM_CONTROL, 1)

        # 5. Program the interrupt thresholds and enable interrupts.
        for c in channels:
            reg = self.gen_s2mm_control_reg(True, other_chan_errors=False, packet_drop=False, bd_complete=True,
                                            timeout=False, error=True, bdcomplete_threshold=2)
            self.write(MCDMA.ADDR_S2MM_CHAN_CONTROL[c], int(reg.hex(), 16))  # writing as bytes leads to corruption

        # 6. Program the TD register of channels. Programming the TDs of a particular channel triggers the fetching of
        # the BDs for the respective channels. The channel pauses when it completes processing for the tail descriptor.
        # it resumes on writing the taildescriptor. If the AXI MCDMA Channel MCDMACR.RS bit is set to 0, a
        # write by the CPU to the TAILDESC_PTR register has no effect except to reposition the pause point.
        for c in channels:
            self.write(MCDMA.ADDR_S2MM_TAILDESC[c], self.bd_chains[c].tail_addr & 0xffffffff)
            self.write(MCDMA.ADDR_S2MM_TAILDESC[c]+4, self.bd_chains[c].tail_addr >> 32)


class S2MMBufferDescriptor:
    """
    S2MM BD:
    next pointer 32bits
        0:5 Must be 0, descriptors must be 16-word aligned that is, 0x00, 0x40, 0x80, and so forth. Any other
            alignment has undefined results.
        6:31 value must be 16word aligned (e.g multiple of 0x40)
    next pointer upper 32
    buffer addr 32bits
        - must be data width aligned if core lacks Buffer Realignment Engine (ours do)
    buffer addr upper 32
    reserved 32bits
    reserved 32bits
    control 32bits
        0-25 This value indicates the amount of space in bytes available for receiving data in an S2MM stream.
            The usable width of buffer length is specified by the parameter Width of Buffer Length Register
            (c_sg_length_width). A maximum of 67,108,863 bytes of transfer can be described by this field. This
            value should be an integer multiple of AXI4-Stream data width. Note: The total buffer space in the S2MM
            descriptor chain (that is, the sum of buffer length values for each descriptor in a chain) must be, at a
            minimum, capable of holding the maximum receive packet size. Undefined results occur if a packet larger
            than the defined buffer space is received. Setting the Buffer Length Register Width smaller than 26
            reduces FPGA resource utilization.
        26  End of Frame. Flag indicating the last buffer to be processed. This flag is set by the sw/user to indicate
            to AXI DMA that this descriptor describes the end of the packet. The buffer associated with this
            descriptor is received last.
            0 = Not End of Frame.
            1 = End of Frame.
            This is applicable only when AXI_DMA is configured in Micro mode.
        27 RXSOF  Start of Frame. Flag indicating the first buffer to be processed. This flag is set by the
            sw/user to indicate to AXI DMA that this descriptor describes the start of the packet. The buffer
            associated with this descriptor is received first.
            0 = Not Start of Frame.
            1 = Start of Frame.
            This is applicable only when AXI_DMA is configured in Micro mode.
        28-31 Reserved, should be set to 0
    status 32bits (set by core)
        0:25 Transferred Bytes This value indicates the amount of data received and stored in the buffer described by
            this descriptor. This might or might not match the buffer length. For example, if this descriptor
            indicates a buffer length of 1,024 bytes but only 50 bytes were received and stored in the buffer,
            then the Transferred Bytes field indicates 0x32. The entire receive packet length can be determined by
            adding the Transferred Byte values from each descriptor from the RXSOF descriptor to the Receive End of
            Frame (RXEOF) descriptor. The usable width of Transferred Bytes is specified by the parameter
            Width of Buffer Length Register (c_sg_length_width) in the block design. A maximum of 67,108,863 bytes of
            transfer can be described by this field.
        26 RXEOF End of Frame. Flag indicating buffer holds the last part of packet. This bit is set by AXI MCDMA to
            indicate that the buffer associated with this descriptor contains the end of the packet. User Application
            data sent through the status stream input is stored in APP0 to APP4 of the RXEOF descriptor when the
            Control/Status Stream is enabled.
        27 RXSOF Start of Frame. Flag indicating buffer holds first part of packet. This bit is set by AXI MCDMA to
            indicate that the buffer associated with this descriptor contains the start of the packet.
        28 DMAIntErr MCDMA Internal Error. Internal Error detected by AXI DataMover. This error can occur if a 0
            length bytes to transfer is fed to the AXI DataMover. This only happens if the Buffer Length specified in
            the fetched descriptor is set to 0. This error can also be caused if there is an under-run or
            over-run condition. This error condition causes the AXI MCDMA to halt gracefully. The MCDMACR.RS bit is
            set to 0, and when the engine has completely shut down, the MCDMASR.Halted bit is set to 1.
        29 DMASlvErr DMA Slave Error. Slave Error detected by primary AXI DataMover. This error occurs if the slave
            read from the Memory Map interface issues a Slave Error. This error condition causes the AXI MCDMA to
            halt gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down, the
            MCDMASR.Halted bit is set to 1.
        30 DMADecErr DMA Decode Error. Decode Error detected by primary AXI DataMover. This error occurs if the
            Descriptor Buffer Address points to an invalid address. This error condition causes the AXI MCDMA to halt
            gracefully. The MCDMACR.RS bit is set to 0, and when the engine has completely shut down,
            the MCDMASR.Halted bit is set to 1.
        31 Complete This indicates that the MCDMA engine has completed the transfer as described by the
            associated descriptor. The MCDMA Engine sets this bit to 1 when the transfer is completed. The software can
            manipulate any descriptor with the Completed bit set to 1. If a descriptor is fetched with this bit set to
            1, the descriptor is considered a stale descriptor. A SGIntErr is flagged and the AXI MCDMA engine halts.
    User0 - User3 32bits each
        When Status/Control Stream is enabled, the status data received on the AXI Status Stream is stored into the
        APP fields of the End of Frame (EOF) Descriptor. For other S2MM descriptors with EOF = 0, the APP fields are
        set to zero by the Scatter Gather Engine.
    User4 32bits
        User Application field 4 and Receive Byte Length. If Use RxLength In Status Stream is not enabled, this field
        functions identically to APP0 to APP3 in that the status data received on the AXI Status Stream is stored into
        the APP4 field of the End of Frame (EOF) Descriptor. This field has a dual purpose when Use RxLength in
        Status Stream is enabled. The first least significant bits specified in the Buffer Length Register Width
        (up to a maximum of 16) specify the total number of receive bytes for a packet that were received on the
        S2MM primary data stream. Second, the remaining most significant bits are User Application data.
    """
    S2MM_STATUS_BD = bitstruct.compile('b1b1 b1b1 b1b1 u26', names=['complete', 'dma_decode_err',
                                                        'dma_slave_err', 'dma_internal_err',
                                                        'rx_sof', 'rx_eof', 'xfer_len'])
    BD = bitstruct.compile('>u64 u64 '
                           'p64'
                           'p4 b1b1 u26'
                           'b1b1 b1b1 b1b1 u26'
                           'u32 u32 u32 u32 u32', names=['next', 'buff',
                                                         'control.rxsof', 'control.rxeof', 'control.buff_len',
                                                         'status.complete', 'status.dma_decode_err',
                                                         'status.dma_slave_err', 'status.dma_internal_err',
                                                         'status.rx_sof', 'status.rx_eof', 'status.xfer_len',
                                                         'app0', 'app1', 'app2', 'app3', 'app4'])

    def __init__(self, next_addr, buf_addr, length, _dict=None):
        """
        Note that this may generate an illegal BD:
            next must be in multiples of 0x40
            Buf_addr mod stream_width_bytes must be 0. E.g. for a 64 byte wide stream it must be in multiples of 0x40
            length must be representable by the width of the core's Buffer Length Register
        """
        # NB Changing the hard coded defaults will set non-zero initial values for BD properties that are set by the
        # DMA core, while this probably has no effect it may make it harder to debug what comes back.
        # setting status.complete=True would initialize the BD as stale.
        defaults = {'next': next_addr, 'buff': buf_addr,
                     'control.rxsof':0, 'control.rxeof':0, 'control.buff_len': length,
                     'status.xfer_len': 0, 'status.rx_eof': False, 'status.rx_sof': False,
                     'status.dma_internal_err': False, 'status.dma_slave_err': False,
                     'status.dma_decode_err': False, 'status.complete': False,
                     'app0': 0, 'app1': 0, 'app2': 0, 'app3': 0, 'app4': 0}
        if _dict:
            defaults.update(_dict)
        self._bytes = self.BD.pack(defaults)

    @staticmethod
    def from_dict(d):
        return S2MMBufferDescriptor(0, 0, 0, _dict=d)

    @staticmethod
    def array_to_dict(array):
        d = {'next': np.uint64(array[0]) | (np.uint64(array[1]) << np.uint64(32)),
             'buff': np.uint64(array[2]) | (np.uint64(array[3]) << np.uint64(32)),
             'control.buff_len': array[6] & ((1 << 26) - 1),
             'status.xfer_len': array[7] & ((1 << 26) - 1),
             'status.rx_eof': bool(array[7] & (1 << 26)), 'status.rx_sof': bool(array[7] & (1 << 27)),
             'status.dma_internal_err': bool(array[7] & (1 << 28)),
             'status.dma_slave_err': bool(array[7] & (1 << 29)),
             'status.dma_decode_err': bool(array[7] & (1 << 30)),
             'status.complete': bool(array[7] & (1 << 31))}
        if array.size>7:
            d.update({'app0': array[8], 'app1': array[9], 'app2': array[10],
                      'app3': array[11], 'app4': array[12]})
        return d

    @staticmethod
    def from_array(array):
        if array.size not in (8, 12, 16):
            raise ValueError('Improper data, 8, 12, or 16 words required')
        return S2MMBufferDescriptor(0, 0, 0,
                                    _dict=S2MMBufferDescriptor.array_to_dict(array[:12]))

    @property
    def array(self):
        """ Return 8 or 12 words with the contents of the buffer descriptor. """
        n=len(self._bytes)//4
        words = [int(self._bytes[i * 4:(i + 1) * 4].hex(), 16) for i in range(n)]
        words[:2] = words[:2][::-1]  # gotta swap so that the high word goes into the right register
        words[2:4] = words[2:4][::-1]
        words.extend([0]*(16-len(words)))
        return np.array(words, dtype=np.uint32)


class S2MMBufferChain:
    def __init__(self, n_buffers, buffer_size, dtype='u8', cyclic=True):
        # if buffer_size % dtype().itemsize:
        #     raise ValueError('Buffer size must be a multiple of the itemsize width (0x40)')
        self._chain = allocate((n_buffers, 16), dtype='u4')
        if self._chain.device_address % 0x40:
            raise ValueError('Chain does not start on a 0x40 address boundary')

        self._buffers = pynq.allocate((n_buffers, buffer_size), dtype=dtype)
        self.chain_length = n_buffers
        item_nbytes= int(dtype[1:])

        self.buf_addr = [self._buffers.device_address + i*buffer_size*item_nbytes for i in range(n_buffers)]
        self.buffer_size = buffer_size
        self._buffers[:] = 0
        self.cyclic=cyclic

        self.chain_addr = [self._chain.device_address + 0x40 * i for i in range(n_buffers)]

        # Build chain
        for i, buf_addr in zip(range(n_buffers), self.buf_addr):
            if i==n_buffers-1:
                next_addr = self.chain_addr[0] if self.cyclic else 0
            else:
                next_addr = self.chain_addr[i+1]
            descriptor = S2MMBufferDescriptor(next_addr, buf_addr, self.buffer_size*item_nbytes)
            self._chain[i, :] = descriptor.array

    @property
    def head_addr(self):
        """ Return the memory address of the first descriptor in the chain"""
        return self._chain[0].device_address

    @property
    def tail_addr(self):
        """ Return the memory address of the last descriptor in the chain"""
        return  self._chain[-1].device_address

    def cyclify_index(self, i):
        if not self.cyclic and i<0 or i>=self.chain_length:
            raise IndexError(f'Index {i} out of bounds for a non-cyclic chain of length {self.chain_length}')
        i%=self.chain_length
        return i

    def descriptor(self, i):
        """Return a dict of link i of the chain"""
        return self._chain[self.cyclify_index(i)]

    def descriptor_dict(self, i):
        """Return a dict of link i of the chain"""
        return S2MMBufferDescriptor.array_to_dict(self._chain[self.cyclify_index(i)])

    def descriptor_data(self, i):
        """Retrieve the data from the descriptor"""
        i = self.cyclify_index(i)

        cd = self._chain[i]
        cd.flush()
        desc=S2MMBufferDescriptor.from_array(cd)

        # sd = unpack_s2mm_status(self._descr[fetch, 7])
        if desc.DMAIntErr:
            assert desc.size != 0, 'a descriptor has null size'
            raise RuntimeError('DMA under/overrun, dropping data??')
        if desc.DMADecErr:
            raise RuntimeError('Invalid buffer descriptor address')
        if desc.DMASlvErr:
            raise RuntimeError('Error on slave read of the Memory Map interface')
        if not desc.complete:
            raise RuntimeError('IOC but descriptor not completed')
        if not desc.RXSOF or not desc.RXEOF:
            raise NotImplementedError('Packet spans multiple descriptors')
        # if not cd[7]>>31:
        #     raise RuntimeError('Descriptor not complete')
        self._buffers[i].flush()
        d = np.array(self._buffers[i, :desc.n_transferred])
        cd[7]&=0x7FFF_FFFF
        return d

    def __del__(self):
        self._buffers.freebuffer()
        self._chain.freebuffer()


class _CyclicS2MMDMAChannel(_SGDMAChannel):
    def __init__(self, existing_rx):
        super().__init__(existing_rx._mmio, existing_rx._max_size, existing_rx._align,
                         existing_rx._tx_rx, existing_rx._dre, interrupt=existing_rx._interrupt)
        self.chain = None

    def dma_info(self, s2mmchain=True):
        S2MM_DMACR = MM2S_DMACR = (  # (name, offset, length)
            ('irq_delay', 24, 8),
            ('irq_threshold', 16, 8),
            ('error_irq_enable', 14, 1),
            ('delay_irq_enable', 13, 1),
            ('complete_irq_enable', 12, 1),
            ('cyclic_bd_enable', 4, 1), ('keyhole', 3, 1), ('reset', 2, 1), ('runstop', 0, 1))

        S2MM_DMASR = MM2S_DMASR = (  # (name, offset, length)
            ('delay_status', 24, 8),
            ('threshold_status', 16, 8),
            ('error_irq', 14, 1),
            ('delay_irq', 13, 1),
            ('complete_irq', 12, 1),
            ('sg_decode_error', 10, 1), ('sg_slave_error', 9, 1), ('sg_internal_error', 8, 1),
            ('dma_decode_error', 6, 1), ('dma_slave_error', 5, 1), ('dma_internal_error', 4, 1),
            ('sg_included', 3, 1), ('idle', 1, 1), ('halted', 0, 1))

        S2MM_CURDESC = MM2S_CURDESC = (('current_descriptor', 0, 64),)
        S2MM_TAILDESC = MM2S_TAILDESC = (('tail_descriptor', 0, 64),)
        S2MM_DA = MM2S_SA = (('address', 0, 64),)
        S2MM_LENGTH = MM2S_LENGTH = (('length', 0, 26),)
        SG_CTL = (('sg_user', 8, 4),
                  ('sg_cache', 0, 4))

        def pack(description, d):
            result = 0
            for n, o, l in description:
                v = d[n]
                result |= (v & (2 ** l - 1)) << o
            return result

        def unpack(description, x):
            result = {}
            for n, o, l in description:
                result[n] = (x >> o) & (2 ** l - 1)
            return result

        regs = (('MM2S_DMACR', 0, 1, MM2S_DMACR),
                ('MM2S_DMASR', 0x4, 1, MM2S_DMASR),
                ('MM2S_CURDESC', 0x8, 2, MM2S_CURDESC),
                ('MM2S_TAILDESC', 0x10, 2, MM2S_TAILDESC),
                ('MM2S_SA', 0x18, 2, MM2S_SA),
                ('MM2S_LENGTH', 0x28, 1, MM2S_LENGTH),
                ('SG_CTL', 0x2c, 1, SG_CTL),
                ('S2MM_DMACR', 0x30, 1, S2MM_DMACR),
                ('S2MM_DMASR', 0x34, 1, S2MM_DMASR),
                ('S2MM_CURDESC', 0x38, 2, S2MM_CURDESC),
                ('S2MM_TAILDESC', 0x40, 2, S2MM_TAILDESC),
                ('S2MM_DA', 0x4c, 2, S2MM_DA),
                ('S2MM_LENGTH', 0x58, 1, S2MM_LENGTH))

        for k, a, l, r in regs:
            x = self._mmio.read(a, l * 4)
            print(k, x, unpack(r, x), '\n')
        if s2mmchain:
            for i in range(self.chain.chain_length):
                print(self.chain.descriptor(i).physical_address, self.chain.descriptor_dict(i))


    @property
    def s2mm_currdesc(self):
        return self._mmio.read(self._offset+8, length=8, word_order='little')

    def s2mm_dmasr(self, unpack=False):
        x= self._mmio.read(self._offset + 4)
        if unpack:
            x = S2MMBufferDescriptor.S2MM_STATUS_BD.unpack(np.array([x]))
        return x

    @property
    def error_str(self):
        error = self.s2mm_dmasr
        if not error & 0x770:  # error bits
            return None
        if error & 0x10:
            return "DMA Internal Error (transfer length 0?)"
        if error & 0x20:
            return "DMA Slave Error (cannot access memory map interface)"
        if error & 0x40:
            return "DMA Decode Error (invalid address)"
        if error & 0x100:
            return "Scatter-Gather Internal Error (re-used completed descriptor)"
        if error & 0x200:
            return "Scatter-Gather Slave Error (cannot access memory map interface)"
        if error & 0x400:
            return "Scatter-Gather Decode Error (invalid descriptor address)"

    def transfer(self):
        """
        Transfer memory with the DMA

        Transfer must only be called when the channel is halted
        For `nbytes`, 0 means everything after the starting point.

        If the AXI DMA is not configured for data re-alignment then a
        valid address must be aligned or undefined results occur.

        For S2MM (recv), if Data Realignment Engine is not included,
        the destination address must be S2MM Memory Map data width aligned.

        For example, if memory map data width = 32, data is aligned if it is
        located at word offsets (32-bit offset), that is, 0x0, 0x4, 0x8, 0xC,
        and so forth.

        Cyclic Buffer Descriptor (BD) mode allows the DMA to loop through the
        buffer descriptors without user intervention. As the DMA cycles through
        the BDs indefinitely, the wait() function is not valid in this mode.
        Instead, use the stop() function to terminate DMA operation. This mode
        is only valid for the sendchannel.

        Parameters
        ----------
        array : ContiguousArray
            An contiguously allocated array to be transferred
        start : int
             Offset into array to start. Default is 0.
        nbytes : int
             Number of bytes to transfer. Default is 0.
        cyclic : bool
             Enable cyclic BD mode. Default is False.

        """
        if not self.halted:
            raise RuntimeError("DMA channel not halted")

        self._cyclic = False

        # Idle DMA engine
        self.stop()

        # Figure out largest possible block size

        blk_size = self._max_size - (self._max_size % self._align)
        n = blk_size//np.uint64().itemsize
        if blk_size%np.uint64().itemsize:
            raise RuntimeError('buffer size is not an appropriate multiple')

        self.chain = S2MMBufferChain(4, dtype='u8', buffer_size=n, cyclic=True)

        start = 0
        if not self._dre and ((self.chain._buffers.physical_address + start) % self._align) != 0:
            raise RuntimeError(
                "DMA does not support unaligned transfers; "
                "Starting address must be aligned to "
                "{} bytes.".format(self._align)
            )

        self.chain._buffers[:]=0
        if self._flush_before:
            self.chain._buffers.flush()

        # Flush DMA descriptors
        self.chain._chain.flush()
        self._descr = self.chain._chain


        desc0 = self.chain.descriptor(0)
        descN = self.chain.descriptor(self.chain.chain_length-1)

        # Write first desc
        self._mmio.write(self._offset + 0x08, desc0.physical_address & 0xFFFFFFFF)
        self._mmio.write(self._offset + 0x0C, (desc0.physical_address>> 32) & 0xFFFFFFFF)

        self._active_buffer = self.chain._buffers

        # Let's go!
        self.transferred = 0
        self._mmio.write(self._offset, 0x1001)  # Interrupt on complete non-cyclic (cyclic: 0x1011
        while not self.running:
            pass

        # Writing last desc triggers the descriptor fetches
        self._mmio.write(self._offset + 0x10, descN.physical_address & 0xFFFFFFFF)
        self._mmio.write(self._offset + 0x14, (descN.physical_address>> 32) & 0xFFFFFFFF)

    def wait(self, use_asyncio=False):
        if use_asyncio:
            try:
                loop = asyncio.get_running_loop()
            except:
                loop = asyncio.new_event_loop()
            task = loop.create_task(self._interrupt.wait())
            loop.run_until_complete(task)
        else:
            while not self.s2mm_dmasr() & 0x7000:  # an interrupt
                pass

    def yield_buf(self, use_asyncio=False):
        self.transfer()

        self.packets = []
        max_packets = 40
        while True:
            self.wait(use_asyncio=use_asyncio)
            status = self.s2mm_dmasr(unpack=True)
            if status['dma_decode_err'] or status['dma_slave_err'] or status['dma_internal_err']:
                error_str = 'Errors: '
                if status['dma_decode_err']:
                    error_str+='dma_decode_err '
                if status['dma_slave_err']:
                    error_str += 'dma_slave_err '
                if status['dma_internal_err']:
                    error_str += 'dma_internal_err '
                raise RuntimeError(error_str)
            elif status['idle'] or status['halted']:
                break
            elif status['complete']:
                i = self.chain.chain_addr.index(self.s2mm_currdesc)
                i = self.chain.cyclify_index(i-1)
                try:
                    d = self.chain.descriptor_data(i)
                except RuntimeError as e:
                    break
                self.s2mm_taildesc = self.chain.chain_addr[i]
                self.packets.append(d)
                if len(self.packets) > max_packets:
                    break
            else:
                raise RuntimeError('Not possible')

    @property
    def s2mm_taildesc(self):
        return self.mmio.read(self._offset+0x10, 8, 'little')

    @s2mm_taildesc.setter
    def s2mm_taildesc(self, x):
        self.mmio.write(self._offset+0x10, x & 0xffffffff)
        self.mmio.write(self._offset+0x14, (x  >> 32) & 0xffffffff)



from logging import getLogger
import numpy as np
import time

import mkidgen3.drivers.rfdcclock
import mkidgen3.opfb as dsp
from . import util

_gen3_overlay, _frequencies = [None] * 2


def set_waveform(freq, amplitudes=None, attenuations=None, simple=False, phases=None, **kwarg):
    """
    Configure the DAC replay table to output a waveform containing the specified frequencies.

    frequencies are in Hz, amplitudes are from 0-1, attenuations are in dB

    Additional keyword arguments are passed on when calling the dac table drivers replay() method.

    Do this either simply, and directly within this function or via either dactable.daccomb or
    dactable.generate_dac_comb.

    Manual loading of a waveform can be attained by performing
    table = generate_dac_comb(frequencies=np.array([0.3e9]), n_samples=2**19, sample_rate=4.096e9,
                              amplitudes=np.array([1.0]))
    ol.dac_table.stop()
    ol.dac_table.replay(table['iq'], fpgen=lambda x: (x*2**15).astype(np.uint16))

    and looking at the contents of ol.dac_table._buffer which should match

    buf = np.zeros((2 ** 15, 2, 16), dtype=np.int16)
    buf[:, 0, :] = table['iq'].real.reshape((2 ** 15,16)) * 2**15
    buf[:, 1,: ] = table['iq'].imag.reshape((2 ** 15,16)) * 2**15

    in Hz amplitude 0-1"""
    from .daccomb import daccomb, generate_dac_comb
    n_samples = 2 ** 19
    sample_rate = 4.096e9
    n_res = 2048

    # my old ipython test channelizer code
    if simple:
        if amplitudes is None:
            amplitudes = np.ones_like(freq)
        t = 2 * np.pi * np.arange(n_samples) / sample_rate
        comb = np.zeros(n_samples, dtype=np.complex64)
        phases = np.zeros(n_res)
        for i in range(freq.size):
            comb += amplitudes[i] * np.exp(1j * (t * freq[i] + phases[i]))
        dactable = {'comb': comb}
    else:
        if attenuations is not None:
            dactable = daccomb(frequencies=freq, n_samples=n_samples, attenuations=attenuations,
                               sample_rate=sample_rate, return_full=True, phases=phases)
            comb = dactable['comb']
        else:
            if amplitudes is None:
                amplitudes = np.ones_like(freq)

            dactable = generate_dac_comb(frequencies=freq, n_samples=n_samples, sample_rate=sample_rate,
                                         amplitudes=amplitudes, phases=phases)
            comb = dactable['iq']

    getLogger(__name__).debug(f"Comb shape: {comb.shape}. \n"
                              f"Total Samples: {comb.size}. Memory: {comb.size * 4 / 1024 ** 2:.0f} MB\n")
    play_waveform(comb, **kwarg)
    return dactable


def play_waveform(iq, **kwargs):
    _gen3_overlay.dac_table.replay(iq, stop_if_needed=True, **kwargs)


def set_channels(freq):
    """
    Set each resonator channel to use the correct OPFB bin given the 2048 frequencies (in Hz) for each channel.

    Only the first 2048 frequencies will be used. Channels are assigned in the order frequencies are specified.
    If fewer than 2048 frequencies are specified no assumptions may be made about which OPFB bin number(s) drive the
    remaining channels, however they may be determined by inspecting the .bins property of bin_to_res.
    """
    freq = freq[:2048]
    bins = np.arange(2048, dtype=int)
    bins[:freq.size] = dsp.opfb_bin_number(freq)
    _gen3_overlay.photon_pipe.reschan.bin_to_res.bins = bins


def iq_find_phase(n_points=1024):
    """A helper function to capture data just after bin2res and after reschan. NEEDS WORK"""
    x = _gen3_overlay.capture.capture_iq(n_points, 'all', tap_location='iq')
    iq = np.array(x)
    x.freebuffer()
    iq = iq[..., 0] + iq[..., 1] * 1j

    phase = np.angle(iq)
    return -phase.mean(0) / (2 * np.pi), iq


def capture_opfb(ol, n=256, raw=False):
    """Capture the OPFB output, exercise caution with large n as the result is copied from PL to PS DDR4"""
    out = np.zeros((n, 4096, 2) if raw else (n, 4096), dtype=np.int16 if raw else np.complex64)
    ol.photon_pipe.reschan.bin_to_res.bins = range(0, 2048)
    x = ol.capture.capture_iq(n, 'all', tap_location='rawiq')
    if raw:
        out[:, :2048, :] = x
        x.freebuffer()
    else:
        out[:, :2048] = util.buf2complex(x, free=True)
    ol.photon_pipe.reschan.bin_to_res.bins = range(2048, 4096)

    x = ol.capture.capture_iq(n, 'all', tap_location='rawiq')
    if raw:
        out[:, 2048:, :] = x
        x.freebuffer()
    else:
        out[:, 2048:] = util.buf2complex(x, free=True)

    return out



from logging import getLogger
import zmq
from mkidgen3.server.feedline_config import IFConfig, FeedlineConfig, BitstreamConfig, RFDCClockingConfig, RFDCConfig, WaveformConfig, ChannelConfig, DDCConfig, FeedlineConfig, FilterConfig, TriggerConfig
from mkidgen3.server.captures import CaptureJob, FRSClient, CaptureRequest, StatusListener
from mkidgen3.server.waveform import WaveformFactory
from mkidgen3.server.feedline_config import FeedlineConfigManager
import numpy as np
from mkidgen3.util import setup_logging
setup_logging('feedlineclient')

frsa = FRSClient(url='mkidrfsoc4x2.physics.ucsb.edu', command_port=8888, data_port=8889, status_port=8890)
frsb = FRSClient(url='rfsoc4x2b.physics.ucsb.edu', command_port=8888, data_port=8889, status_port=8890)

# Bitstream Config
bitstream = BitstreamConfig(bitstream='/home/xilinx/gen3_top_final.bit', ignore_version=True)
# RFDC Clocking Config
rfdc_clk = RFDCClockingConfig(programming_key='4.096GSPS_MTS_dualloop', clock_source='external')
# RDFC Config
rfdc = RFDCConfig(dac_mts=True, adc_mts=False, adc_gains=None, dac_gains=None)
# IF Config
if_board = IFConfig(lo=6000, adc_attn=50, dac_attn=50)

#feedline config
fc = FeedlineConfig(bitstream=bitstream, rfdc_clk=rfdc_clk, rfdc=rfdc,
                    if_board=if_board, waveform=WaveformConfig(), chan=ChannelConfig(), ddc=DDCConfig(),
                    filter=FilterConfig(), trig=TriggerConfig())

one_gib_adc = 2**28
four_gib_adc = 4*2**28

size = four_gib_adc
tap = 'adc'
cr = CaptureRequest(n=4*2**28, tap=tap, feedline_config=fc, feedline_server=frsa)
j = CaptureJob(cr)
j.submit(True, True)
result = j.data(timeout=300)
print('hi')



import numpy as np
from typing import Iterable

DAC_MAX_OUTPUT_DBM = 1  # [dBm] see Xilinx DS926
DAC_MAX_INT = 8191  # see PG269 p.219
ADC_RESOLUTION = 14 # bits see PG269 p.219
ADC_DAC_INTERFACE_WORD_LENGTH = 16  # bits see PG269 p.219
# This computes the max int the ADC can register, For the RFSOC4x2 this is 0x7FFC, which is given in the table titled
# "RF-ADC and RF-DAC Word Interface" in PG269 under section "Digital Data Format"
ADC_MAX_INT = 2**(ADC_RESOLUTION-1)-1 << (ADC_DAC_INTERFACE_WORD_LENGTH-ADC_RESOLUTION) # see above
ADC_INPUT_WARN = 0.7*ADC_MAX_INT
ADC_MAX_INPUT_DBM = 1  # [dBm] see Xilinx DS926

# per DS926 Table RF-ADC Electrical Characteristics for ZU4xDR Devices
# (https://docs.xilinx.com/r/en-US/ds926-zynq-ultrascale-plus-rfsoc/RF-ADC-Electrical-Characteristics)
ADC_MAX_V = 0.7
ADC_MAX_V_DIFF = 1.4
ADC_MAX_V_CONSERVATIVE = 0.4
ADC_MAX_V_DIFF_CONSERVATIVR = 0.8

DAC_RESOLUTION = 14  # bits see PG269 p.219
DAC_LUT_SIZE = 2 ** 19  # values
ADC_SAMPLE_RATE = 4.096e9  # Hz
DAC_SAMPLE_RATE = 4.096e9  # Hz
DAC_FREQ_RES = DAC_SAMPLE_RATE / DAC_LUT_SIZE
DAC_FREQ_MAX = (DAC_SAMPLE_RATE / 2) - DAC_FREQ_RES
DAC_FREQ_MIN = -DAC_SAMPLE_RATE / 2
N_OPFB_CHANNELS = 4096  # Number of OPFB channels
N_CHANNELS = 2048  # Number of DDC (resonator) channels
N_PHASE_GROUPS = 128  # See HLS and the phase drivers, the AXI streams are 16 channels wide
N_IQ_GROUPS = 256  # See HLS and the iq drivers, the AXI streams are 8 channels wide
N_POSTAGE_CHANNELS = 8  # see HLS and the postage filter driver
SYSTEM_BANDWIDTH = 4.096e9  # Hz Full readout bandwidth
OS = 2  # OPFB Overlap factor
OPFB_CHANNEL_SAMPLE_RATE = OS * ADC_SAMPLE_RATE / N_OPFB_CHANNELS


# PHASE_IQ_INPUT_FRACTIONAL_BITS = 14  #commented as no clear meaning, no doc, and no usage in codebase

# cordic output scaled radians [-1,1] is 18 bits, signed, 3 integer, 15 fractional, truncated to the low 16 bits
PHASE_FRACTIONAL_BITS = 15

PL_TOTAL_BYTES = 4 * 1024 ** 3
SYSTEM_OVERHEAD_BYTES = 2 * 768 * 1024 ** 2
COMPRESSION_OVERHEAD_BYTES = 128 * 1024 ** 2  #this is complete speculation
LOWPASSED_IQ_SAMPLE_RATE = 1e6
MAXIMUM_DESIGN_COUNTRATE_PER_S = 5000

PHOTON_POSTAGE_WINDOW_LENGTH = 127  # Must be 1 less than the HLS value, this is the number of IQ values captured for a photon event


from mkidgen3.equipment_drivers.ifboard import MAX_IN_ATTEN, IF_ATTN_STEP, MAX_OUT_ATTEN


def channel_to_iqgroup(channels: Iterable | int) -> set:
    """
    Convert channels to IQ groups which contains those channels.
    Args:
        channels: resonator channels (0-2047)

    Returns: set of IQ groups

    """
    try:
        return set([g // 8 for g in channels])
    except TypeError:
        return {channels//8}


def iqgroup_to_channel(iq_group: Iterable | int) -> set:
    """
    Convert IQ groups to channels.
    Args:
        iq_group: (0-128)

    Returns: set of corresponding channels (0-2047)

    """
    chan_per_group = N_CHANNELS // N_IQ_GROUPS
    if isinstance(iq_group, int):
        return set(np.arange(chan_per_group)+iq_group*chan_per_group)
    else:
        iq_group = np.fromiter(iq_group, int, len(iq_group))
        return set((np.arange(chan_per_group)+(iq_group*chan_per_group)[:, np.newaxis]).flatten())


def channel_to_phasegroup(channels: Iterable | int) -> set:
    """
    Convert phase channels to IQ groups which contains those channels.
    Args:
        channels: resonator channels (0-2047)

    Returns: set of IQ groups (0-255)

    """
    try:
        return set([g // 16 for g in channels])
    except TypeError:
        return {channels//16}


def phasegroup_to_channel(phase_group: Iterable | int) -> set:
    """
    Convert phase groups to channels.
    Args:
        phase_group: (0-127)

    Returns: set of corresponding channels (0-2047)

    """
    chan_per_group = N_CHANNELS // N_PHASE_GROUPS
    if isinstance(phase_group, int):
        return set(np.arange(chan_per_group)+phase_group*chan_per_group)
    else:
        phase_group = np.fromiter(phase_group, int, len(phase_group))
        return set((np.arange(chan_per_group)+(phase_group*chan_per_group)[:, np.newaxis]).flatten())



from pynq import DefaultIP
from mkidgen3.system_parameters import DAC_LUT_SIZE, DAC_SAMPLE_RATE, PL_DDS_M
import logging

class SweepDDS(DefaultIP):
    """
    A class to interface with the re-programmable DDS implmented in the programmable logic (PL DDS).
    The PL DDS is able to generate frequencies between the DAC table resolution (7.8125 kHz) and the 
    input clock (128 MHz) with <DAC table resoluton> precision though in practice it performs best 
    at lower frequencies (<3 MHz).


    Attributes:
        see DefaultIP

    Methods:
        configure()
            Program PL DDS.

    Example:
       pl_dds = ol.freq_sweep.PL_DDS.dds_pinc_1
       pl_dds.configure(7812.5)
    """

    def __init__(self, description):
        super().__init__(description=description)

    bindto = ['slaclab:hls:dds_pinc:1.0']

    def configure(self, freq):
        """
        Program DDS to output a given frequency.

        Args:
            freq: Desired PL DDS output frequency in Hz. Allowed Values: (7812.5 to 128e6).
        """

        fres = DAC_SAMPLE_RATE/DAC_LUT_SIZE
        max_freq = fres*PL_DDS_M
        
        if not ((fres <= freq <= max_freq) or freq == 0): # check freq
            raise ValueError(f"DDS frequency must be either 0 or between {fres} Hz and {max_freq} Hz.")
        
        pinc = int(freq/(fres)) & 0x3FFF # quantize & ensure 14-bit expected by DDS IP
        
        # log messages
        logging.getLogger(__name__).info(f"Programming PL DDS with {pinc*fres} Hz")
        logging.getLogger(__name__).debug(f"Setting phase increment to {pinc}.")
        
        self.write(0x10, pinc)
        return
    
    def off(self):
        """
        Turn DDS off by making it output a constant 1 (identity) that doesn't up or downconvert anything.
        """
        self.configure(0)
